from __future__ import print_function

from pprint import pprint
from vmad.core.operator import operator, ZeroGradient
from vmad.core.model import Builder
import pytest

@operator
class error_on_grad:
    ain = 'x'
    aout = 'y'

    def apl(node, x):
        return x

    def vjp(node, _y):
        raise AssertionError("shall not reach here")

    def jvp(node, x_):
        raise AssertionError("shall not reach here")

@operator
class error:
    ain = 'x'
    aout = 'y'

    def apl(node, x):
        raise AssertionError("shall not reach here")

    def vjp(node, _y):
        raise AssertionError("shall not reach here")

    def jvp(node, x_):
        raise AssertionError("shall not reach here")

def test_operator_zero():

    with Builder() as m:
        a = m.input('a')
        t1 = error_on_grad(x=a)
        m.output(c=t1)

    init = dict(a=3)

    c, tape = m.compute(init=init, vout='c', return_tape=True)
    assert c == 3

    vjp = tape.get_vjp()
    init = dict(_c=ZeroGradient)
    _a = vjp.compute(init=init, vout='_a', monitor=print)
    assert _a == 0

    jvp = tape.get_jvp()
    init = dict(a_=ZeroGradient)
    c_ = jvp.compute(init=init, vout='c_', monitor=print)
    assert c_ == 0

def test_operator_multiple():
    @operator
    class with_defaults:
        ain = [('x', '*')]
        aout = 'y'

        def apl(node, x): return x
        def vjp(node, _y): return _y
        def jvp(node, x_): return x_

    @operator
    class with_defaults:
        ain = [('x', '*'), 'y']
        aout = 'y'

        def apl(node, x, y): return x
        def vjp(node, _y): return _y, _y
        def jvp(node, x_, y_): return x_, y_

    @operator
    class with_defaults:
        ain = 'x', 'y'
        aout = 'y'

        def apl(node, x, y): return x
        def vjp(node, _y): return _y, _y
        def jvp(node, x_, y_): return x_, y_


def test_operator_defaults():
    @operator
    class with_defaults:
        ain = 'x'
        aout = 'y'

        def apl(node, x, defaults=False):
            assert defaults == False
            return x

        def vjp(node, _y, defaults=False):
            assert defaults == False
            return _y

        def jvp(node, x_, defaults=False):
            assert defaults == False
            return x_

    with Builder() as m:
        a = m.input('a')
        t1 = with_defaults(x=a)
        m.output(c=t1)

    init = dict(a=3)

    c, tape = m.compute(init=init, vout='c', return_tape=True)
    assert c == 3

    vjp = tape.get_vjp()
    init = dict(_c=1)
    _a = vjp.compute(init=init, vout='_a', monitor=print)
    assert _a == 1

    jvp = tape.get_jvp()
    init = dict(a_=1)
    c_ = jvp.compute(init=init, vout='c_', monitor=print)
    assert c_ == 1

def test_operator_skip_unused():

    with Builder() as m:
        a = m.input('a')
        t1 = error(x=a)
        m.output(c=a)

    init = dict(a=3)

    c, tape = m.compute(init=init, vout='c', return_tape=True)
    assert c == 3

    vjp = tape.get_vjp()
    init = dict(_c=0)
    _a = vjp.compute(init=init, vout='_a', monitor=print)
    assert _a == 0

    jvp = tape.get_jvp()
    init = dict(a_=0)
    c_ = jvp.compute(init=init, vout='c_', monitor=print)
    assert c_ == 0

import numpy
@operator
class split:
    ain = 'x'
    aout = 'args'

    def apl(node, x, axis):
        return [numpy.take(x, i, axis=axis) for i in range(numpy.shape(x)[axis])]

    def vjp(node, _args, axis):
        return numpy.stack(_args, axis=axis)

    def jvp(node, x_, axis):
        return [numpy.take(x_, i, axis=axis) for i in range(numpy.shape(x_)[axis])]

@operator
class stack:
    ain = 'args'
    aout = 'y'

    def apl(node, args, axis):
        return numpy.stack(args, axis=axis)

    def vjp(node, _y, args, axis):
        return [numpy.take(_y, i, axis=axis) for i in range(numpy.shape(_y)[axis])]

    def jvp(node, args_, args, axis):
        return numpy.stack(args_, axis)

@operator
class zero_jac:
    ain = 'args'
    aout = 'y'

    def apl(node, args):
        return dict(y=3, nargs=len(args))

    def vjp(node, nargs):
        return ZeroGradient

    def jvp(node, nargs):
        return ZeroGradient



def test_operator_list_in():
    from numpy.testing import assert_array_equal

    with Builder() as m:
        a = m.input('a')
        t = stack(args=[a, a, a], axis=1)
        m.output(c=t)

    init = dict(a=[1, 2])

    c, tape = m.compute(init=init, vout='c', return_tape=True)
    assert_array_equal(c, [[1, 1, 1], [2, 2, 2]])

    vjp = tape.get_vjp()
    init = dict(_c=[[1, 1, 1], [1, 1, 1]])
    _a = vjp.compute(init=init, vout='_a', monitor=print)

    assert_array_equal(_a, [3, 3])

    jvp = tape.get_jvp()
    init = dict(a_=[1, 1])
    c_ = jvp.compute(init=init, vout='c_', monitor=print)

    assert_array_equal(c_, [[1, 1, 1], [1, 1, 1]])

# This was based on VHBoehm and Maxelee's test case on stdlib.eval
# failure with a list input in MADLens
def test_operator_list_with_zero_vjp():
    with Builder() as m:
        value = m.input('value')
        r = zero_jac([value])
        m.output(my_array=r)

    init = dict(value=1)
    my_array, tape = m.compute(init=init, vout='my_array', return_tape=True)
    assert my_array == 3

    vjp = tape.get_vjp()
    _value = vjp.compute(init=dict(_my_array=5), vout='_value')
    assert _value == 0

    jvp = tape.get_jvp()
    my_array_= jvp.compute(init=dict(value_=5), vout='my_array_')
    assert my_array_ == 0

def test_operator_list_out():
    from numpy.testing import assert_array_equal
    from vmad.core.symbol import List, Symbol

    with Builder() as m:
        a = m.input('a')

        # it is very awkward to prepare a list output
        # I doubt this will be of any usefulness
        b1 = Symbol('b1')
        b2 = Symbol('b2')
        l = List([b1, b2])

        t = split(x=a, axis=0, args=l)
        assert isinstance(t, List)
        assert next(iter(t)) is b1

        m.output(c=l)

    init = dict(a=[[1, 1], [2, 2]])

    c, tape = m.compute(init=init, vout='c', return_tape=True, monitor=print)
    assert_array_equal(c, [[1, 1], [2, 2]])

    vjp = tape.get_vjp()
    init = dict(_c=[[1, 1], [1, 1]])
    _a = vjp.compute(init=init, vout='_a', monitor=print)

    assert_array_equal(_a, [[1, 1], [1, 1]])

    jvp = tape.get_jvp()
    init = dict(a_=[[1, 1], [1, 1]])
    c_ = jvp.compute(init=init, vout='c_', monitor=print)

    assert_array_equal(c_, [[1, 1], [1, 1]])


def test_operator_multi_out():
    @operator
    class op:
        ain = 'x'
        # for python 2.x need to use this syntax
        # to preserve orders
        aout = 'y1', 'y2'

        def apl(node, x):
            return dict(y1=x, y2=2 * x)
        def vjp(node, _y1, _y2):
            return dict(_x = _y1 + 2 * _y2)
        def jvp(node, x_):
            return dict(y1_=x_, y2_=2 * x_)

    with Builder() as m:
        a = m.input('a')
        t1, t2 = op(x=a)
        m.output(c=t1, d=t2)

    init = dict(a=3)

    (c, d), tape = m.compute(init=init, vout=('c', 'd'), return_tape=True)
    assert c == 3
    assert d == 6

    vjp = tape.get_vjp()
    init = dict(_c=1, _d=1)
    _a = vjp.compute(init=init, vout='_a', monitor=print)
    assert _a == 3

    jvp = tape.get_jvp()
    init = dict(a_=1)
    c_, d_ = jvp.compute(init=init, vout=('c_', 'd_'), monitor=print)
    assert c_ == 1
    assert d_ == 2

def test_operator_default_jvp():
    @operator
    class op:
        ain = 'x'
        aout = 'y1'

        def apl(node, x):
            return dict(y1=x * 2)
        def vjp(node, _y1):
            return dict(_x = _y1 * 2)

    with Builder() as m:
        a = m.input('a')
        t1 = op(x=a)
        m.output(c=t1)

    init = dict(a=3)

    c, tape = m.compute(init=init, vout='c', return_tape=True)
    assert c == 6

    vjp = tape.get_vjp()
    init = dict(_c=1)
    _a = vjp.compute(init=init, vout='_a', monitor=print)
    assert _a == 2

    jvp = tape.get_jvp()
    init = dict(a_=1)
    c_ = jvp.compute(init=init, vout=('c_'), monitor=print)
    assert c_ == 2


def test_operator_multi_out_unused():
    @operator
    class op:
        ain = 'x'
        # for python 2.x need to use this syntax
        # to preserve orders
        aout = 'y1', 'y2'

        def apl(node, x):
            return dict(y1=x, y2=2 * x)
        def vjp(node, _y1, _y2):
            return dict(_x = _y1 + 2 * _y2)
        def jvp(node, x_):
            return dict(y1_=x_, y2_=2 * x_)

    with Builder() as m:
        a = m.input('a')
        t1, t2 = op(x=a)
        m.output(c=t1)

    init = dict(a=3)

    c, tape = m.compute(init=init, vout='c', return_tape=True)
    assert c == 3

    vjp = tape.get_vjp()
    init = dict(_c=1)
    _a = vjp.compute(init=init, vout='_a', monitor=print)
    assert _a == 1

    jvp = tape.get_jvp()
    init = dict(a_=1)
    c_ = jvp.compute(init=init, vout=('c_'), monitor=print)
    assert c_ == 1

def test_operator_record():
    # assert used extra args are recored on the tape
    @operator
    class myrecord:
        ain = 'x'
        aout = 'y'

        def apl(node, x, p):
            return x * p

        def rcd(node, x, p, y):
            return dict(x=x, u=p)

        def vjp(node, x, _y, u):
            return _y * u

        def jvp(node, x, x_, u):
            return x_ * u

    with Builder() as m:
        a = m.input('a')
        b = myrecord(x=a, p=2.0)
        m.output(b=b)

    init = dict(a = 1.0)
    b, tape = m.compute(init=init, vout='b', monitor=print, return_tape=True)

    assert b == 2.0
    assert 'p' not in tape[0].impl_kwargs
    assert 'u' in tape[0].impl_kwargs

def test_operator_record_extra():
    # assert used extra args are recored on the tape
    @operator
    class myrecord:
        ain = 'x'
        aout = 'y'

        def apl(node, x, p):
            return dict(y=x * p, extra=p)

        def rcd(node, x, p, y, extra):
            return dict(x=x, u=p, extra=extra)

        def vjp(node, x, _y, u):
            return _y * u

        def jvp(node, x, x_, u):
            return x_ * u

    with Builder() as m:
        a = m.input('a')
        b = myrecord(x=a, p=2.0)
        m.output(b=b)

    init = dict(a = 1.0)
    b, tape = m.compute(init=init, vout='b', monitor=print, return_tape=True)

    assert b == 2.0
    assert 'p' not in tape[0].impl_kwargs
    assert 'u' in tape[0].impl_kwargs




from vmad import operator

import numpy

# a few commonly used operators are expected to be
# compatible with the python ones.
from vmad.core.stdlib import mul, add, abs, pow, div, mod

# import all functions defined in unary module
from vmad.lib.unary import *

from numpy.core.einsumfunc import _parse_einsum_input

@operator
class broadcast_to:
    ain = 'x'
    aout = 'y'
    def apl(node, x, shape):
        return dict(y=numpy.broadcast_to(x, shape), xshape=numpy.shape(x), yshape=shape)

    def vjp(node, _y, xshape, yshape):
        _x = numpy.broadcast_to(_y, yshape)

        # remove all newly added axes
        if len(yshape) != len(xshape):
            newaxis = tuple(range(0, len(yshape) - len(xshape)))
            _x = numpy.sum(_x, axis=newaxis)

            yshape = yshape[len(yshape) - len(xshape):]

        for axis, size in enumerate(xshape):
            if size == 1: # broadcasted axis
                _x = numpy.sum(_x, axis=axis, keepdims=True)

        return dict(_x=_x)


def _join_einsum_sub(sub_op, sub_y):
    return ','.join(sub_op) + '->' + sub_y

@operator
class einsum:
    ain = 'x'
    aout = 'y'
    def apl(node, subscripts, x):
        operands = []
        operands.append(subscripts)
        operands.extend(x)
        # we use the internal einsum function to parse subscripts to a normal form.
        sub_op, sub_y, operands = _parse_einsum_input(operands)
        sub_op = sub_op.split(',')
        return dict(y=numpy.einsum(_join_einsum_sub(sub_op, sub_y), *x), sub_op=sub_op, sub_y=sub_y, x=x)

    def vjp(node, x, _y, sub_op, sub_y):
        _x = []
        for i, arg in enumerate(x):
            # Jacobian is removing the item from the chain. 
            # vjp is replace the item with _y.
            a = list(sub_op)[:]
            x1 = list(x)[:]
            new_sub_y = a[i]
            a[i] = sub_y
            x1[i] = _y
            _x.append(numpy.einsum(_join_einsum_sub(a, new_sub_y), *x1))

        return _x

    def jvp(node, x, x_, sub_op, sub_y):
        y_ = []
        for i, arg in enumerate(x):
            # Jacobian is removing the item from the chain. 
            # jvp is replace the item with x_
            x1 = list(x)[:]
            x1[i] = x_[i]
            y_.append(numpy.einsum(_join_einsum_sub(sub_op, sub_y), *x1))

        # sum of all branches
        return numpy.sum(y_, axis=0)

def matmul(x, A):
    return einsum('ik,ijk->ij', x, A)

@operator
class unpack_complex:
    ain = {'x' : '*'}
    aout = [('real', '*'), ('imag', '*')]

    def apl(node, x):
        return dict(real=x.real, imag=x.imag)

    def vjp(node, _real, _imag):
        return dict(_x = _real + _imag * 1j)

@operator
class pack_complex:
    ain = [('real', '*'), ('imag', '*')]
    aout = {'y' : '*'}

    def apl(node, real, imag):
        return dict(y = real + imag * 1j)

    def vjp(node, _y):
        return dict(_real = _y.real, _imag = _y.imag)

@operator
class conj:
    ain = 'x'
    aout = 'y'

    def apl(node, x):
        return numpy.conj(x)

    def vjp(node, _y):
        return numpy.conj(_y)

@operator
class to_scalar:
    ain  = {'x': 'ndarray'}
    aout = {'y': '*'}

    def apl(node, x):
        return dict(y = (x * numpy.conj(x)).sum())

    def vjp(node, _y, x):
        return dict(_x = 2 * numpy.conj(_y) * x)

    def jvp(node, x_, x):
        return dict(y_ = (x_ * numpy.conj(x) + numpy.conj(x_) * x).sum())

@operator
class copy:
    ain = {'x' : 'ndarray'}
    aout = {'y' : 'ndarray'}

    def apl(node, x):
        return dict(y = numpy.copy(x))

    def vjp(node, _y):
        return dict(_x = numpy.copy(_y))

@operator
class stack:
    ain = {'x' : 'ndarray',}
    aout = {'y' : 'ndarray'}

    def apl(node, x, axis):
        return dict(y=numpy.stack(x, axis=axis))

    def vjp(node, _y, axis):
        return dict(_x=[numpy.take(_y, i, axis=axis)
                for i in range(numpy.shape(_y)[axis])])

@operator
class take:
    ain = {'x' : 'ndarray',}
    aout = {'y' : 'ndarray'}

    def apl(node, x, i, axis):
        if axis is None:
            raise AssertionError('Assertion error. axis keyword in linalg.take cannot be None.')
        i = numpy.array(i, dtype='intp')
        y = numpy.take(x, i, axis=axis)
        yshape = numpy.shape(y)
        xshape = numpy.shape(x)
        return dict(y=y, yshape=yshape, xshape=xshape, i=i, axis=axis)

    def vjp(node, _y, i, axis, xshape, yshape):
        # shape of y is x.shape [..., i
        _x = numpy.zeros(xshape)
        _y = numpy.broadcast_to(_y, yshape)

        indices = [slice(None)] * len(xshape)
        indices[axis] = i

        # one element can show up several times, add all gradients.
        numpy.add.at(_x, tuple(indices), _y)
        return dict(_x=_x)

@operator
class concatenate:
    ain = {'x' : '*',}
    aout = {'y' : 'ndarray'}

    def apl(node, x, axis):
        if axis is None:
            raise AssertionError('Assertion error. axis keyword in linalg.take cannot be None.')
        return dict(y=numpy.concatenate(x, axis=axis))

    def rcd(node, x, axis, y):
        return dict(xshapes=[numpy.shape(x1) for x1 in x], axis=axis)

    def vjp(node, _y, axis, xshapes):
        # chopping of _y along the axis will do
        _y = numpy.swapaxes(_y, 0, axis)
        offset = 0
        _x = []

        for xshape in xshapes:
            _x1 = _y[offset:offset+xshape[axis]]
            _x.append(numpy.swapaxes(_x1, 0, axis))
            offset = offset + xshape[axis]

        return dict(_x=_x)

@operator
class transpose:
    ain = {'x' : 'ndarray',}
    aout = {'y' : 'ndarray'}

    def apl(node, x, axes):
        return dict(y=numpy.transpose(x, axes))

    def rcd(node, x, axes, y):
        return dict(axes=axes)

    def vjp(node, _y, axes):
        inverse = [0] * len(axes)
        for i, a in enumerate(axes):
            inverse[a] = i
        _x = _y.transpose(inverse)
        return dict(_x=_x)

@operator
class reshape:
    ain  = {'x' : '*'}
    aout = {'y': '*'}

    def apl(node, x, shape):
        return dict(y = numpy.reshape(x, shape))

    def rcd(node, x, shape, y):
        return dict(xshape = numpy.shape(x), shape=shape)

    def vjp(node, _y, xshape):
        return dict(_x=_y.reshape(xshape))

@operator
class sumat:
    ain  = {'x' : '*'}
    aout = {'y': '*'}

    def apl(node, x, at, axis=0):
        if not (numpy.diff(at) >= 0).all():
            raise ValueError('at must be monotonically increasing')

        y = numpy.add.reduceat(x, at, axis=axis, dtype='f8')
        return y

    def rcd(node, x, y, at, axis=0):
        return dict(xshape = numpy.shape(x), at=at, axis=axis)

    def vjp(node, _y, xshape, at, axis):
        _x = numpy.ones(xshape)
        N = numpy.diff(numpy.concatenate([at, [xshape[axis]]], axis=0))
        return numpy.repeat(_y, N, axis=axis)

@operator
class sum:
    ain  = {'x' : '*'}
    aout = {'y': '*'}

    def apl(node, x, axis=None):
        return dict(y = numpy.sum(x, axis=axis, dtype='f8'))

    def rcd(node, x, y, axis=None):
        return dict(xshape = numpy.shape(x), axis=axis)

    def vjp(node, _y, xshape, axis):
        _x = numpy.ones(xshape)

        if axis is not None:
            # prepend to the correct axis
            _yshape = list(numpy.shape(_y))
            _yshape.insert(axis, 1)
            _y = _y.reshape(_yshape)

        _x *= _y
        return dict(_x = _x)




class ModelError(Exception): pass

class BadArgument(ModelError): pass
class UnpackError(ModelError): pass
class DuplicatedOutput(ModelError): pass
class MissingArgument(ModelError): pass
class OverwritePrecaution(ModelError): pass
class UnexpectedOutput(ModelError): pass
class ResolveError(ModelError): pass
class InferError(ModelError): pass
class BrokenPrimitive(ModelError): pass

class ExecutionError(ModelError):
    def __init__(self, msg, reason):
        self.reason = reason
        ModelError.__init__(self, msg)

def makeExecutionError(msg, reason):
    errortype = type("ExecutionError(%s)" % type(reason).__name__,
            (type(reason), ExecutionError), {})
    return errortype(msg, reason)



class Node:
    """ A node on the computing graph.

        The node is the first argument to apl(node, ....) and
        jvp / vjp functions.

        node[argname] gives the input symbol
    """
    def __init__(self, primitive, _frameinfo):
        self.primitive = primitive
        self.operator = primitive.operator
        self.prototype = primitive.operator.prototype
        self._frameinfo = _frameinfo

        # add a few aliases for accessing primitive attributes
        # 
        self.name = primitive.name

        self._varin = {} # references
        self._varout = {}

    def __getitem__(self, key):
        """ getting input variables as symbols """
        # varin are references.
        return self._varin[key].symbol

    @property
    def varin(self):
        return self._varin

    @property
    def varout(self):
        return self._varout

    def __repr__(self):
        #return "%s(%s=>%s) at %s:%d" % (type(self).__name__, self.varin, self._varout, self._frameinfo[0], self._frameinfo[1])
        return "%s @ %s : %s " % (self.name, self._frameinfo[0], self._frameinfo[1])

    def call(self, kwargs):
        """ call the implementation function of the primitive;

            invoked by the Context

            kwargs: the arguments that goes into the impl function

            Returns: dict, result for each varout.
        """
        from .symbol import BaseSymbol
        for key, value in kwargs.items():
            assert not isinstance(value, BaseSymbol)

        r = self.primitive.impl(self, **kwargs)

        # allow returning without using a dict
        # if there is only a single output argument
        if not isinstance(r, dict):
            if len(self.varout) == 1:
                argname = next(iter(self.varout.keys()))
                r = {argname:r}
            if len(self.varout) == 0:
                if r is not None:
                    raise ValueError("Return value of the primitive is not None, while no output arguments are defined")
                r = {}

        for key, value in r.items():
            assert not isinstance(value, BaseSymbol)

        return r

    def record(self, kwargs, r):
        """ generate the kwargs that goes into the tape;
            default is to record the entire kwargs.

            Sometimes we do not need the entire kwargs; e.g.
            for linear operators we only need enough information to create
            the output array of the back-prop gradient
            but we don't need the actual parameters.

            invoked by the Context.

            kwargs: the arguments that goes into the impl function
            r : the result of the calculation operator apl, dict from argname to value
                see above.

            Returns: dict that goes into the tape, will be available in vjp and jpv
        """
        # merge the two dictionaries, prioritizing kwargs (inputs).
        d = {}
        d.update(r)
        d.update(kwargs)
        from .symbol import BaseSymbol
        for key, value in d.items():
            assert not isinstance(value, BaseSymbol)
        return self.primitive.record_impl(self, **d)

    def find_primitive_type(node, func):
        # we will only do this on the apl primitives
        # because otherwise this is undefined
        # the algebra of autodiff in vmad3 is explicitly not closed!
        assert node.primitive == node.operator.apl

        assert func in ['vjp', 'jvp', 'apl']

        if func == 'jvp': return node.operator.jvp
        if func == 'vjp': return node.operator.vjp
        if func == 'apl': return node.operator.apl

    def is_literal(self, argname):
        from vmad.core.symbol import Literal
        return isinstance(self.varin[argname].symbol, Literal)




def example():
    graph = []

    graph.append(add(x1='a', x2='b', y='c'))

    c, tape = compute(graph, vout='c', init=dict(a=10, b=20), return_tape=True)

from vmad.core.registry import Registry

def test_registry():
    reg = Registry()

    reg.register("add***", dict(x="*", y="*", z='*'), name='add')
    reg.register("add**", dict(x="*", y="*"), name='add')
    reg.register("addii", dict(x=int, y=int), name='add')
    reg.register("addi*", dict(x=int, y="*"), name='add')
    reg.register("add*i", dict(x="*", y=int), name='add')


    assert reg.match("add", dict(x=1, y=2., z="12")) == 'add***'
    assert reg.match("add", dict(x=1, y=2.)) == 'addi*'
    assert reg.match("add", dict(x=1, y=1 )) == 'addii'
    assert reg.match("add", dict(x="s", y=1 )) == 'add*i'
    assert reg.match("add", dict(x="s", y="s" )) == 'add**'








from __future__ import print_function

from vmad import Builder
from vmad import autooperator
from vmad import operator

from vmad.lib.linalg import add
from pprint import pprint

def test_vmad3_functional():
    """ this test demonstrates building a model directly """
    with Builder() as m:
        a, b = m.input('a', 'b')

        t1 = add(a, a)
        t2 = add(b, 0)
        c = add(t1, t2)

        m.output(c=c)

    print("----- model -----")
    pprint(m)
    pprint(m[:])

    print("----- compute -----")
    init = dict(a=3, b=4)

    c = m.compute(init=init, vout='c')
    print(init, c)

    print("----- tape -----")
    init = dict(a=3, b=4)
    c, tape = m.compute(init=init, vout='c', return_tape=True)
    print(init, c)
    pprint(tape)

    print("----- vjp -----")
    vjp = tape.get_vjp()
    pprint(vjp)
    pprint(vjp[:])

    init = dict(_c=1.0)
    _a, _b = vjp.compute(init=init, vout=['_a', '_b'], monitor=print)
    print('_a, _b = ', _a, _b)

    print("----- jvp -----")
    jvp = tape.get_jvp()
    pprint(jvp)
    pprint(jvp[:])

    init = dict(a_=1.0, b_=1.0)
    c_, = jvp.compute(init=init, vout=['c_'], monitor=print)
    print('c_ = ', c_)

def test_autooperator():
    """ this test demonstrates using autooperator
        to define an operator based on a model, whose
        vjp and jvp are obtained via autodiff.
    """
    @autooperator
    class mymodel:
        ain = {'a' : '*',
               'b' : '*'}
        aout = {'c' : '*'}

        def main(model, a, b, n):
            for i in range(n):
                a = add(a, a)

            t2 = add(b, 0)
            return dict(c=add(a, t2))

    init = dict(a=3, b=4)

    m3 = mymodel.build(n=3)
    print("----- model 2-----")
    pprint(m3)
    c = m3.compute(init=init, vout='c')
    assert c == 3 * 2 ** 3 + 4

    m2 = mymodel.build(n=2)
    print("----- model 3-----")
    pprint(m2)
    c = m2.compute(init=init, vout='c')
    assert c == 3 * 2 ** 2 + 4

    # complicated model is longer
    assert len(m3[:]) > len(m2[:])




from vmad import operator
from vmad.core.symbol import Literal, ZeroLiteral
from vmad.core.symbol import Symbol
import numpy

class unary_ufunc:
    ain = 'x'
    aout = 'y'

    @staticmethod
    def f(x): return x

    @staticmethod
    def fprime(x): return 1

    def apl(node, x):
        return node.operator.prototype.f(x)

    def vjp(node, x, _y):
        return node.operator.prototype.fprime(x) * _y

    def jvp(node, x, x_):
        return node.operator.prototype.fprime(x) * x_

@operator
class sinc(unary_ufunc):
    @staticmethod
    def f(x): return numpy.sinc(x)
    @staticmethod
    def fprime(x):
        # FIXME: expand this properly:
        y = 1 / x * (
          numpy.cos(numpy.pi * x)
        - numpy.sin(numpy.pi * x) / (numpy.pi * x))
        return numpy.where(x == 0, 0, y)

@operator
class sin(unary_ufunc):
    @staticmethod
    def f(x): return numpy.sin(x)
    @staticmethod
    def fprime(x): return numpy.cos(x)

@operator
class cos(unary_ufunc):
    @staticmethod
    def f(x): return numpy.cos(x)
    @staticmethod
    def fprime(x): return -numpy.sin(x)

@operator
class tan(unary_ufunc):
    @staticmethod
    def f(x): return numpy.tan(x)
    # FIXME: hard code a faster version which saves tan.
    @staticmethod
    def fprime(x): return 1 + numpy.tan(x) ** 2

@operator
class arcsin(unary_ufunc):
    @staticmethod
    def f(x): return numpy.arcsin(x)
    @staticmethod
    def fprime(x): return 1 / (1 - x **2) **0.5

@operator
class arccos(unary_ufunc):
    @staticmethod
    def f(x): return numpy.arccos(x)
    @staticmethod
    def fprime(x): return -1 / (1 - x **2) **0.5

@operator
class arctan(unary_ufunc):
    @staticmethod
    def f(x): return numpy.arctan(x)
    @staticmethod
    def fprime(x): return 1 / (1 + x **2)

@operator
class sinh(unary_ufunc):
    @staticmethod
    def f(x): return numpy.sinh(x)
    @staticmethod
    def fprime(x): return numpy.cosh(x)

@operator
class cosh(unary_ufunc):
    @staticmethod
    def f(x): return numpy.cosh(x)
    @staticmethod
    def fprime(x): return numpy.sinh(x)


@operator
class log(unary_ufunc):
    @staticmethod
    def f(x): return numpy.log(x)
    @staticmethod
    def fprime(x): return 1.0 / x

@operator
class log10(unary_ufunc):
    @staticmethod
    def f(x): return numpy.log10(x)
    @staticmethod
    def fprime(x, log10=1.0 / numpy.log(10)): return 1.0 / x * log10

@operator
class exp(unary_ufunc):
    @staticmethod
    def f(x): return numpy.exp(x)
    @staticmethod
    def fprime(x): return numpy.exp(x)

@operator
class absolute(unary_ufunc):
    @staticmethod
    def f(x): return numpy.absolute(x)
    @staticmethod
    def fprime(x): return 1.0 * (x > 0) + -1.0 * (x < 0)

fabs = absolute



"""
Vocaburary
----------

universal : a variable whose value is replicated on all ranks.
distributed : a variable whose value is partitioned on all ranks.
root-only   : a variable whose value is only stored on the root rank.

collective : an operation that is performed by all ranks.

"""

from vmad import operator, autooperator
import numpy

@operator
class allreduce:
    ain = [('x', '*')]
    aout = [('y', '*')]

    def apl(node, x, comm):
        return comm.allreduce(x)

    def vjp(node, _y, comm):
        return dict(_x = _y)

    def jvp(node, x_, comm):
        return dict(y_ = comm.allreduce(x_))

@operator
class allbcast:
    """
    The invert operator of allreduce.

    Converting an universal variable 'x' into a distributed variable 'x'.

    The input variable is assumed to be identical
    on all ranks. Therefore in the forward pass the allbcast operator
    is an nop operation.

    Unlike broadcast which converts a master variable into a universal
    variable.

    """

    ain = [('x', '*')]
    aout = [('y', '*')]

    def apl(node, x, comm):
        # value should already be identical on all ranks.
        comm.barrier()
        return x

    def vjp(node, _y, comm):
        return dict(_x = comm.allreduce(_y))

    def jvp(node, x_):
        return dict(y_ = x_)





from .error import UnexpectedOutput, makeExecutionError, ModelError
from .stdlib import terminal
from collections import OrderedDict

_raise_internal_errors = True

def set_raise_internal_errors(flag):
    """ If raise_internal_errors is set to True, then the errors in
        node execution are directly raised.

        If False, we will produce a wrapped messsage,
        which contains the line number where the
        node is declared. It is easier for debugging and in Python 3+
        the underlying error message is also printed.
    """
    global _raise_internal_errors
    _raise_internal_errors = flag


class Context(dict):
    """ A context is a collection of python objects referred by symbol names;

        This is where the execution of a model occurs.

        Context is the internal API. Use the compute method of a model instead.
    """
    def __init__(self, **init):
        self.update(init)

    def remove_unused(self, nodes):
        """ remove objects not used by nodes"""
        used = set()
        for p in nodes:
            for argname, ref in p.varin.items():
                used = used.union(ref.get_symbol_names())

        toremove = []
        for key in self:
            if not key in used:
                toremove.append(key)

        for key in toremove:
            self.pop(key)

    @classmethod
    def result_used(self, node):
        # FIXME: this doesn't remove all of the unused
        # may need to fix this in 'compile' or 'optimize'.
        if node.primitive == terminal.apl:
            return True

        for argname, var in node.varout.items():
            if var._has_reference(): return True

        if len(node.varout) == 0:
            # special operators without any output always gets run.
            return True
        return False

    def compute(self, model, vout, tape, monitor=None):
        """
            compute a model in the current context (self)
        """
        _voutnames = set([var._name for var in model._vout])

        for varname in vout:
            if varname not in _voutnames:
                raise UnexpectedOutput("Requested vout %s is not defined by the model as an output; available ones are %s" % (varname, _voutnames))

        r = {}
        for i, node in enumerate(model):

            if self.result_used(node):
                self.execute(node, tape)

            if node.primitive == terminal.apl:
                for argname, var in node.varout.items():
                    r[var._name] = self[var._name]

            self.remove_unused(model[i+1:])

            if monitor is not None:
                monitor(node, node.varin, node.varout, self)

        r = OrderedDict([(varname, r[varname]) for varname in vout])

        return r

    def execute(self, node, tape):
        """ execute a node on the context, recording the
            arguments of the impl to the tape for replay, debuggin,
            or autodiff.
        """

        resolved = {}
        for argname, ref in node.varin.items():
            var = ref.symbol
            resolved[argname] = var._resolve(self)

        kwargs = {}
        kwargs.update(resolved)

        if _raise_internal_errors:
            r = node.call(kwargs)
        else:
            try:
                r = node.call(kwargs)
            except Exception as e:
                raise makeExecutionError(
                    "Error computing node : %s" % (node), e)

        if tape is not None:
            tape.append(node, node.record(kwargs, r))

        for argname, var in node.varout.items():
            var._store(self, r[argname])




from vmad import Builder, autooperator
from vmad.lib import mpi, linalg

from pmesh.pm import ParticleMesh
import numpy

@autooperator
class MPIChiSquareOperator:
    ain = [('x', '*')]
    aout = [('y', '*')]

    def main(self, x, comm):
        chi2 = linalg.sum(x * x)
        chi2 = mpi.allreduce(chi2, comm)
        return dict(y = chi2)

from abopt.abopt2 import Problem as BaseProblem, VectorSpace

class MPIVectorSpace(VectorSpace):
    def __init__(self, comm):
        self.comm = comm

    def addmul(self, a, b, c, p=1):
        """ a + b * c ** p, follow the type of b """
        if p is not 1: c = c ** p
        c = b * c
        if a is not 0: c = c + a
        return c

    def dot(self, a, b):
        """ einsum('i,i->', a, b) """
        return self.comm.allreduce(numpy.sum(a * b))

class MPIChiSquareProblem(BaseProblem):
    """ Defines a chisquare problem, which is

        .. math::

            y = \sum_i [R_i(F(s))]^2

        F : forward_operator

        [R_i] : residuals

        comm : The problem is defined on a MPI communicator -- must be consistent
        with forward_operator and residuals. (usually just pm.comm or MPI.COMM.WORLD)

    """
    def __init__(self, comm, forward_operator, residuals, vectorspace=None):
        self.residuals = residuals
        self.forward_operator = forward_operator
        self.comm = comm

        if vectorspace is None:
            vectorspace = MPIVectorSpace(comm)

        with Builder() as m:
            x = m.input('x')
            y = 0
            fx = forward_operator(x)
            # fixme: need a way to directly include a subgraphs
            # rather than building it again.
            for operator in self.residuals:
                r = operator(*fx)
                chi2 = MPIChiSquareOperator(r, comm)
                y = linalg.add(y, chi2)
            m.output(y=y)

        def objective(x):
            return m.compute(vout='y', init=dict(x=x))

        def gradient(x):
            y, [vjp] = m.compute_with_vjp(init=dict(x=x), v=dict(_y=1.0))
            return vjp

        def hessian_vector_product(x, v):
            Dv = 0

            replay = forward_operator.precompute(x=x)

            for operator in self.residuals:
                with Builder() as m:
                    xx = m.input('x')
                    fx = replay(xx)
                    r = operator(*fx)
                    m.output(y=r)

                y, [Dv1] = m.compute_with_gnDp(vout='y',
                            init=dict(x=x),
                            v=dict(x_=v))
                Dv = Dv + Dv1
            # H is 2 JtJ, see wikipedia on Gauss Newton.
            return Dv * 2

        BaseProblem.__init__(self,
                        vs = vectorspace,
                        objective=objective,
                        gradient=gradient,
                        hessian_vector_product=hessian_vector_product)

class EpochScheduler(dict):
    """ An object to schedule hyper parameters in multi-epoch
        training.
    """
    def __init__(self):
        pass

    def __call__(self, argname, values):
        self.schedule(argname, values)

    @property
    def min_nepochs(self):
        return max(list(self.keys()) + [0]) + 1

    def schedule(self, argname, values):
        """ schedule a training parameter `argname`
            at the given epochs.

            values can be either a list or a dict, or a scalar.

            list : one value per epoch
            dict : step function per epoch, key is the epoch to change
            scalar : at the 0th epoch.

            The nameed optimizer parameter will be set to the
            given value on that epoch as if all epochs are played
            sequentially.
        """
        if isinstance(values, (list, tuple)):
            iter = enumerate(values)
        elif isinstance(values, (dict,)):
            iter = values.items()
        else:
            # scalar
            iter = enumerate([values])

        for epoch, value in iter:
            args = self.get(epoch, {})
            args[argname] = value
            self[epoch] = args

    def get_args(self, epoch):
        """ move the trainer to the epoch"""
        args = {}
        for epochi in sorted(self.keys()):
            if epochi <= epoch:
                for argname, value in self[epochi].items():
                    args[argname] = value

        return args

class Epoch(object):
    def __init__(self, epoch, nepochs):
        self.epoch = epoch
        self.nepochs = nepochs
    def __repr__(self):
        return "Epoch %d / %d" % (self.epoch, self.nepochs)

class MAPInversion:
    """ Inversion of a ChiSquareProblem with a ForwardOperator;
        given data finding the optimal reconstruction.

        The inversion is done in multiple epochs to accelerate
        convergence and avoid local attractors. Conceptually this
        is similar to stochastic gradient descent.

        Two schedulers are involved. schedule_optimizer controls
        the parameters of the trainer (e.g., maxiter, atol, rtol, etc).
        schedule_problem controls the parameters of the problem
        (subsampling, changing prior, etc).
    """
    def __init__(self, optimizer, ForwardOperator):
        self.optimizer = optimizer
        self.ForwardOperator = ForwardOperator
        self.schedule_optimizer = EpochScheduler()
        self.schedule_problem = EpochScheduler()

    @classmethod
    def create_like(kls, other, epochs=None):
        """ create a new inversion that is a copy of other.
            notice that some of the attributes may be lost;
            not supposed to add new attributes.
        """
        self = object.__new__(kls)
        self.__dict__.update(other.__dict__)

        self.ForwardOperator = other.ForwardOperator
        self.optimizer = other.optimizer
        self.schedule_optimizer = EpochScheduler()
        self.schedule_problem = EpochScheduler()
        if epochs is None:
            self.schedule_optimizer.update(other.schedule_optimizer)
            self.schedule_problem.update(other.schedule_problem)
        else:
            # only take the selected epochs
            for i, epoch in enumerate(epochs):
                self.schedule_optimizer[i] = other.schedule_optimizer.get_args(epoch)
                self.schedule_problem[i] = other.schedule_problem.get_args(epoch)
        return self

    def apply(self, d, s0, epochs=None, monitor_epoch=None, monitor_progress=None):
        """ Apply MAP inversion on synthetic data.

            returns the MAP estimation of the signal.

            epochs : a list of epochs to run; if None, run all of them.

        """
        if epochs is None:
            epochs = range(self.nepochs)

        s1 = s0
        for epoch in epochs:
            if monitor_epoch:
                monitor_epoch(Epoch(epoch, max(epochs) + 1))

            problem = self.get_problem(d, epoch)

            self.optimizer.__dict__.update(self.schedule_optimizer.get_args(epoch))

            state = self.optimizer.minimize(problem, s1, monitor=monitor_progress)
            s1 = state['x']
        return s1

    @property
    def nepochs(self):
        return max(self.schedule_optimizer.min_nepochs,
                     self.schedule_problem.min_nepochs)

    def get_problem(self, d, epoch=0):
        """ Create a problem to be solved for the epoch """
        # if you see exception here, see add a schedule_problem for the arg.
        return self.problem_factory(d=d, **self.schedule_problem.get_args(epoch))

    def problem_factory(self, d, **args):
        """ Create a problem object.

            override this method in subclasses; must
            accept at least the data to be inverted.

            args are set by

            >> mapinversion.schedule_problem(argname, argvalue)
        """
        raise NotImplementedError




from vmad import operator, autooperator
from pmesh.pm import ParticleMesh
from vmad.lib import linalg
import numpy

@operator
class to_scalar:
    ain = {'x' : 'RealField'}
    aout = {'y' : '*'}

    def apl(node, x):
        return dict(y = x.cnorm())

    def vjp(node, _y, x):
        return dict(_x= x * (2 * _y))

    def jvp(node, x_, x):
        return dict(y_ = x.cdot(x_) * 2)

@operator
class as_complex_field:
    ain = {'x' : '*'}
    aout = {'y' : 'ComplexField'}

    def apl(node, x, pm):
        y = pm.create(type='complex')
        y.real[...] = x[..., 0]
        y.imag[...] = x[..., 1]
        return dict(y=y)

    def vjp(node, _y, pm):
        _x = numpy.stack([_y.real, _y.imag], axis=-1)
        return dict(_x=_x)

    def jvp(node, x_, pm):
        y_ = pm.create(type='complex')
        y_.real[...] = x_[..., 0]
        y_.imag[...] = x_[..., 1]
        return dict(y_=y_)

@operator
class r2c:
    ain = {'x' : 'RealField'}
    aout = {'y' : 'ComplexField'}

    def apl(node, x):
        return dict(y=x.r2c(), pm=x.pm)
    def vjp(node, _y, pm):
        _y = pm.create(type='complex', value=_y)
        return dict(_x=_y.r2c_vjp())
    def jvp(node, x_, pm):
        x_ = pm.create(type='real', value=x_)
        return dict(y_=x_.r2c())

@operator
class c2r:
    ain = {'x' : 'ComplexField'}
    aout = {'y' : 'RealField'}

    def apl(node, x):
        return dict(y=x.c2r(), pm=x.pm)
    def vjp(node, _y, pm):
        _y = pm.create(type='real', value=_y)
        return dict(_x=_y.c2r_vjp())
    def jvp(node, x_, pm):
        x_ = pm.create(type='complex', value=x_)
        return dict(y_=x_.c2r())

@operator
class apply_transfer:
    ain = {'x' : 'ComplexField'}
    aout = {'y' : 'ComplexField'}

    def apl(node, x, tf, kind='wavenumber'):
        filter = lambda k, v: v * tf(k)
        return dict(y=x.apply(filter, kind=kind))

    def vjp(node, _y, tf, kind):
        filter = lambda k, v: v * numpy.conj(tf(k))
        return dict(_x=_y.apply(filter, kind=kind))

    def jvp(node, x_, tf, kind):
        filter = lambda k, v: v * tf(k)
        return dict(y_=x_.apply(filter, kind=kind))

def _take_default(array, ind, conj, default):
    ind1 = ind.clip(0, len(array) - 1)
    r = array[ind1]
    r[conj] = numpy.conj(r)[conj]
    # out of bound values, set to default
    numpy.putmask(r, ind != ind1, default)
    return r

@operator
class apply_digitized:
    """
    Apply a digitized transfer function, represented by
    `tf` and a digitizer function, which converts the mode label
    to indices in `tf`.

    if ind < 0 or ind >= len(tf), the mode is skipped.

    y = x * sum Pi lambda_i

    where lambda is the transfer function per binned k, Pi is
        projection operator for binned k.

    Parameters
    ----------
    tf : array_like
        values of digitized transfer function

    digitizer : func([k0, k1, k2, ...]) -> ind, conj
        decided which bin the mode falls in, and if the mode needs
        to be conjugated.
        the mode labels depends on the value of kind.

    kind : string, 'wavenumber', 'circular', 'index'.

    Notes
    -----
    vjp:
        _lambda_i = <x * Pi , _y>
        this is the binned cross power between x and _y.
        for harmonic functions this is always real.
        _x = _y * sum Pi lambda_i

    jvp:
        y_ = x sum Pi lambda_i_ + x_ sum Pi lambda_i

    """
    ain = 'x', 'tf'
    aout = 'y'

    @staticmethod
    def isotropic_wavenumber(kedges):
        def digitizer(k):
            k2 = k.normp(2)
            ind = numpy.digitize(k2 ** 0.5, kedges) - 1
            return ind, numpy.zeros_like(ind, dtype='?')
        return digitizer

    def apl(node, x, tf, digitizer, kind='wavenumber', mode='amplitude'):
        assert mode in ('amplitude', 'phase')

        # build the projection operators, stored as ind, where value of ind is the index in tf.
        ind = numpy.zeros(x.value.shape, dtype='intp')
        conj = numpy.zeros(x.value.shape, dtype='?')

        x.apply(lambda k, v: digitizer(k)[0], kind=kind, out=ind)
        x.apply(lambda k, v: digitizer(k)[1], kind=kind, out=conj)

        # if ind is out of bound from tf, then we return 1
        # because the transfer function is constant 1 on those bins.
        if mode == 'amplitude':
            y = x * _take_default(tf, ind, conj, 1)
            eit = None
        elif mode == 'phase':
            eit = numpy.exp(1j * tf)
            y = x * _take_default(eit, ind, conj, 1)
        return dict(y=y, ind=ind, conj=conj, eit=eit)

    def vjp(node, _y, x, tf, ind, conj, eit, mode):
        ind1 = ind.clip(0, len(tf) - 1)
        mask = ind1 != ind

        if mode == 'amplitude':
            # compute the cross power between _y and x
            weights = _y * numpy.conj(x)
            # count how many modes each value in the field contributes
            weights = weights.apply(x._expand_hermitian, kind='index', out=Ellipsis)
            # clear out of bound modes
            numpy.putmask(weights.value, mask, 0)
            _tf_real = numpy.bincount(ind1.flat, weights=weights.real.flat, minlength=len(tf))
    #        _tf_imag = numpy.bincount(ind1.flat, weights=weights.imag.flat, minlength=len(tf))
            _tf = _tf_real # assuming harmonic functions
            _x = _y * _take_default(tf, ind, conj, 1)

        elif mode == 'phase':
            y = x * _take_default(eit, ind, conj, 0)

            # compute the cross power between _y and x
            weights = _y * numpy.conj(1j * y)
            # count how many modes each value in the field contributes
            weights = weights.apply(x._expand_hermitian, kind='index', out=Ellipsis)
            # clear out of bound modes
            numpy.putmask(weights.value, mask, 0)

            _tf = numpy.bincount(ind1.flat, weights=weights.real.flat, minlength=len(tf))
            _x = _y * _take_default(eit.conj(), ind, conj, 1)

        # gather from other ranks
        _tf = x.pm.comm.allreduce(_tf)

        return dict(_tf = _tf, _x = _x)

    def jvp(node, tf_, x_, x, tf, ind, conj, eit, mode):
        y_ = 0
        if mode == 'amplitude':
            if tf_ is not 0:
                y_ = y_ + x * _take_default(tf_, ind, conj, 0)
            if x_ is not 0:
                y_ = y_ + x_ * _take_default(tf, ind, conj, 1)
        elif mode == 'phase':
            if tf_ is not 0:
                y_ = y_ + x * _take_default(eit * 1j * tf_, ind, conj, 0)
            if x_ is not 0:
                y_ = y_ + x_ * _take_default(eit, ind, conj, 1)

        return y_

@operator
class paint:
    aout = {'mesh' : 'RealField'}
    ain =  {'x': 'ndarray',
           'layout': 'Layout',
           'mass' : 'ndarray'
           }

    def apl(node, x, mass, layout, pm):
        mesh = pm.paint(x, mass=mass, layout=layout, hold=False)
        return dict(mesh=mesh)

    def vjp(node, _mesh, x, mass, layout, pm):
        _mesh = pm.create(type='real', value=_mesh)
        _x, _mass = pm.paint_vjp(_mesh, x, layout=layout, mass=mass)
        return dict(
            _layout = 0,
            _x=_x,
            _mass=_mass)

    def jvp(node, x_, x, layout, mass, layout_, mass_, pm):
        if x_ is 0: x_ = None
        if mass_ is 0: mass_ = None # force cast it to a scalar 0, so make it None
        mesh_ = pm.paint_jvp(x, v_mass=mass_, mass=mass, v_pos=x_, layout=layout)

        return dict(mesh_=mesh_)

@operator
class readout:
    aout = {'value' : 'ndarray'}

    ain = {'x': 'ndarray',
        'mesh': 'RealField',
      'layout' : 'Layout'}

    def apl(node, mesh, x, layout, resampler=None):
        N = mesh.pm.comm.allreduce(len(x))
        value = mesh.readout(x, layout=layout, resampler=resampler)
        return dict(value=value, pm=mesh.pm)

    def vjp(node, _value, x, layout, mesh, pm, resampler=None):
        _mesh, _x = mesh.readout_vjp(x, _value, layout=layout, resampler=resampler)
        return dict(_mesh=_mesh, _x=_x, _layout=0)

    def jvp(node, x_, mesh_, x, layout, layout_, mesh, pm, resampler=None):
        if mesh_ is 0: mesh_ = None
        if x_ is 0: x_ = None
        mesh = pm.create(type='real', value=mesh)
        value_ = mesh.readout_jvp(x, v_self=mesh_, v_pos=x_, layout=layout, resampler=resampler)
        return dict(value_=value_)

@operator
class decompose:
    aout={'layout' : 'Layout'}
    ain={'x': 'ndarray'}

    def apl(node, x, pm):
        return dict(layout=pm.decompose(x))

    def vjp(engine, _layout):
        return dict(_x=0)

    def jvp(engine, x_):
        return dict(layout_=0)

@operator
class gather:
    aout='y'
    ain ='x', 'layout'

    def apl(node, x, layout):
        return dict(y=layout.gather(x))

    def vjp(engine, _y, layout):
        _x = layout.exchange(_y)
        return dict(_layout=0, _x=_x)

    def jvp(engine, layout, layout_, x_,):
        return dict(y_=layout.gather(x_))

@operator
class exchange:
    aout='y'
    ain ='x', 'layout'

    def apl(node, x, layout):
        return dict(y=layout.exchange(x))

    def vjp(engine, _y, layout):
        _x = layout.gather(_y)
        return dict(_layout=0, _x=_x)

    def jvp(engine, layout, layout_, x_,):
        return dict(y_=layout.exchange(x_))

def fourier_space_neg_gradient(dir, pm, order):
    if order == 0:
        nyquist =  numpy.pi / pm.BoxSize[dir] * (pm.Nmesh[dir] - 0.01)
        def filter(k):
            mask = abs(k[dir]) < nyquist
            return - 1j * k[dir] * mask
    elif order == 1:
        def filter(k):
            cellsize = (pm.BoxSize[dir] / pm.Nmesh[dir])
            w = k[dir] * cellsize

            a = 1 / (6.0 * cellsize) * (8 * numpy.sin(w) - numpy.sin(2 * w))
            # a is already zero at the nyquist to ensure field is real
            return (-1j * a)
    else:
        raise ValueError("only order 1 and 2 are supproted")
    return filter

def fourier_space_laplace(k):
    k2 = k.normp(2)
    bad = k2 == 0
    k2[bad] = 1
    k2 = - 1 / k2
    k2[bad] = 0
    return k2

@autooperator('rhok,q->dx1')
def lpt1(rhok, q, pm):
    p = apply_transfer(rhok, fourier_space_laplace)

    layout = decompose(q, pm)

    r1 = []
    for d in range(pm.ndim):
        dx1_c = apply_transfer(p, fourier_space_neg_gradient(d, pm, order=1))
        dx1_r = c2r(dx1_c)
        dx1 = readout(dx1_r, q, layout)
        r1.append(dx1)

    dx1 = linalg.stack(r1, axis=-1)

    return dict(dx1 = dx1)

@autooperator('rhok->rho_lpt2')
def lpt2src(rhok, pm):
    if pm.ndim != 3:
        raise ValueError("LPT 2 only works in 3d")

    D1 = [1, 2, 0]
    D2 = [2, 0, 1]

    potk = apply_transfer(rhok, fourier_space_laplace)

    Pii = []
    for d in range(pm.ndim):
        t = apply_transfer(potk, fourier_space_neg_gradient(d, pm, order=1))
        Pii1 = apply_transfer(t, fourier_space_neg_gradient(d, pm, order=1))
        Pii1 = c2r(Pii1)
        Pii.append(Pii1)

    source = None
    for d in range(pm.ndim):
        source1 = Pii[D1[d]] * Pii[D2[d]]
        if source is None:
            source = source1
        else:
            source = source + source1

    for d in range(pm.ndim):
        t = apply_transfer(potk, fourier_space_neg_gradient(D1[d], pm, order=1))
        Pij1 = apply_transfer(t, fourier_space_neg_gradient(D2[d], pm, order=1))
        Pij1 = c2r(Pij1)
        source1 = - Pij1 * Pij1
        source = source + source1

    source = (3.0 / 7) * source

    return dict(rho_lpt2=source)

@autooperator('wnk->c')
def induce_correlation(wnk, powerspectrum, pm):
    def tf(k):
        k = sum(ki ** 2 for ki in k) ** 0.5
        return (powerspectrum(k) / pm.BoxSize.prod()) ** 0.5

    c = apply_transfer(wnk, tf)
    return dict(c = c)

@autooperator('rhok->dx1,dx2')
def lpt(rhok, q, pm):

    dx1 = lpt1(rhok, q, pm)
    source2 = lpt2src(rhok, pm)
    rhok2 = r2c(source2)
    dx2 = lpt1(rhok2, q, pm)

    return dict(dx1=dx1, dx2=dx2)


@autooperator('dx->f,potk')
def gravity(dx, q, pm):
    from vmad.core.stdlib import watchpoint
    x = q + dx
    #def w(q): print('q', q)
    #watchpoint(x, w)
    #watchpoint(x, lambda x: print('x', q))
    #watchpoint(dx, lambda dx: print('dx', dx))
    layout = decompose(x, pm)

    xl = exchange(x, layout)
    rho = paint(xl, 1.0, None, pm)

    # convert to 1 + delta
    fac = 1.0 * pm.Nmesh.prod() / pm.comm.allreduce(len(q))

    rho = rho * fac
    rhok = r2c(rho)

    p = apply_transfer(rhok, fourier_space_laplace)

    r1 = []
    for d in range(pm.ndim):
        dx1_c = apply_transfer(p, fourier_space_neg_gradient(d, pm, order=1))
        dx1_r = c2r(dx1_c)
        dx1l = readout(dx1_r, xl, None)
        dx1 = gather(dx1l, layout)
        r1.append(dx1)

    f = linalg.stack(r1, axis=-1)
    return dict(f=f, potk=p)

def KickFactor(pt, ai, ac, af):
    return 1 / (ac ** 2 * pt.E(ac)) * (pt.Gf(af) - pt.Gf(ai)) / pt.gf(ac)

def DriftFactor(pt, ai, ac, af):
    return 1 / (ac ** 3 * pt.E(ac)) * (pt.Gp(af) - pt.Gp(ai)) / pt.gp(ac)

@autooperator('dx_i,p_i->dx,p,f')
def leapfrog(dx_i, p_i, q, stages, pt, pm):

    Om0 = pt.Om0

    dx = dx_i
    p = p_i
    f, potk = gravity(dx, q, pm)

    for ai, af in zip(stages[:-1], stages[1:]):
        ac = (ai * af) ** 0.5

        # kick
        dp = f * (KickFactor(pt, ai, ai, ac) * 1.5 * Om0)
        p = p + dp

        # drift
        ddx = p * DriftFactor(pt, ai, ac, af)
        dx = dx + ddx

        # force
        f, potk = gravity(dx, q, pm)

        # kick
        dp = f * (KickFactor(pt, ac, af, af) * 1.5 * Om0)
        p = p + dp

    f = f * (1.5 * Om0)
    return dict(dx=dx, p=p, f=f)

@autooperator('rhok->dx,p')
def firststep(rhok, q, a, pt, pm):

    dx1, dx2 = lpt(rhok, q, pm)

    E = pt.E(a)
    D1 = pt.D1(a)
    D2 = pt.D2(a)
    f1 = pt.f1(a)
    f2 = pt.f2(a)

    dx1 = dx1 * D1
    dx2 = dx2 * D2

    p1 = dx1 * (a ** 2 * f1 * E)
    p2 = dx2 * (a ** 2 * f2 * E)

    p = p1 + p2
    dx = dx1 + dx2
    return dict(dx=dx, p=p)

@autooperator('rhok->dx,p,f')
def nbody(rhok, q, stages, cosmology, pm):
    from fastpm.background import MatterDominated

    stages = numpy.array(stages)
    mid = (stages[1:] * stages[:-1]) ** 0.5
    support = numpy.concatenate([mid, stages])
    support.sort()
    pt = MatterDominated(cosmology.Om0, a=support)

    dx, p = firststep(rhok, q, stages[0], pt, pm)

    dx, p, f = leapfrog(dx, p, q, stages, pt, pm)

    return dict(dx=dx, p=p, f=f)

@operator
class cdot:
    ain = {'x1' : 'ComplexField', 'x2' : 'Complexfield'}
    aout = {'y' : '*'}

    # only keep the real part, assuming two fields are hermitian.
    def apl(node, x1, x2):
        return dict(y=x1.cdot(x2).real)

    def vjp(node, x1, x2, _y):
        _x1 = x2.cdot_vjp(_y)
        _x2 = x1.cdot_vjp(_y)
        return dict(_x1=_x1, _x2=_x2)

    def jvp(node, x1_, x2_, x1, x2):
        return dict(y_=x1.cdot(x2_).real + x2.cdot(x1_).real)

class FastPMSimulation:
    def __init__(self, stages, cosmology, pm, B=1, q=None):
        from fastpm.background import MatterDominated

        if q is None:
            q = pm.generate_uniform_particle_grid()

        stages = numpy.array(stages)
        mid = (stages[1:] * stages[:-1]) ** 0.5
        support = numpy.concatenate([mid, stages])
        support.sort()
        pt = MatterDominated(cosmology.Om0, a=support)
        self.stages = stages
        self.pt = pt
        self.pm = pm
        self.fpm = ParticleMesh(Nmesh=pm.Nmesh * B,
                        BoxSize=pm.BoxSize,
                        dtype=pm.dtype,
                        comm=pm.comm,
                        resampler=pm.resampler)
        self.q = q

    def KickFactor(self, ai, ac, af):
        pt = self.pt
        return 1 / (ac ** 2 * pt.E(ac)) * (pt.Gf(af) - pt.Gf(ai)) / pt.gf(ac)

    def DriftFactor(self, ai, ac, af):
        pt = self.pt
        return 1 / (ac ** 3 * pt.E(ac)) * (pt.Gp(af) - pt.Gp(ai)) / pt.gp(ac)

    @autooperator('rhok->dx, p')
    def firststep(self, rhok):
        q = self.q
        pm = self.pm
        pt = self.pt
        a = self.stages[0]

        dx1, dx2 = lpt(rhok, q, pm)

        E = pt.E(a)
        D1 = pt.D1(a)
        D2 = pt.D2(a)
        f1 = pt.f1(a)
        f2 = pt.f2(a)

        dx1 = dx1 * D1
        dx2 = dx2 * D2

        p1 = dx1 * (a ** 2 * f1 * E)
        p2 = dx2 * (a ** 2 * f2 * E)

        p = p1 + p2
        dx = dx1 + dx2
        return dict(dx=dx, p=p)

    @autooperator('dx->f,potk')
    def gravity(self, dx):
        q = self.q
        # use Force PM resolution including B factor.
        pm = self.fpm

        x = q + dx

        layout = decompose(x, pm)

        xl = exchange(x, layout)
        rho = paint(xl, 1.0, None, pm)

        # convert to 1 + delta
        fac = 1.0 * pm.Nmesh.prod() / pm.comm.allreduce(len(q))

        rho = rho * fac
        rhok = r2c(rho)

        p = apply_transfer(rhok, fourier_space_laplace)

        r1 = []
        for d in range(pm.ndim):
            dx1_c = apply_transfer(p, fourier_space_neg_gradient(d, pm, order=1))
            dx1_r = c2r(dx1_c)
            dx1l = readout(dx1_r, xl, None)
            dx1 = gather(dx1l, layout)
            r1.append(dx1)

        f = linalg.stack(r1, axis=-1)
        return dict(f=f, potk=p)

    @autooperator('rhok->dx,p,f')
    def run(self, rhok):

        dx, p = self.firststep(rhok)

        pt = self.pt
        stages = self.stages
        q = self.q

        Om0 = pt.Om0

        f, potk = self.gravity(dx)

        for ai, af in zip(stages[:-1], stages[1:]):
            ac = (ai * af) ** 0.5

            # kick
            dp = f * (self.KickFactor(ai, ai, ac) * 1.5 * Om0)
            p = p + dp

            # drift
            ddx = p * self.DriftFactor(ai, ac, af)
            dx = dx + ddx

            # force
            f, potk = self.gravity(dx)

            # kick
            dp = f * (self.KickFactor(ac, af, af) * 1.5 * Om0)
            p = p + dp

        f = f * (1.5 * Om0)
        return dict(dx=dx, p=p, f=f)





from .error import ResolveError
import weakref

from . import stdlib

class BaseSymbol(object):
    """ A symbol for building models.

        A symbol is named.

        A symbol can be resolved to a python object
        given a context in respect to its name.

        A symbol may be referenced by multiple

        operators if they take the symbol as an input.

        Only named symbols are registered for autodiff.
    """
    def __init__(self):
        self._parent = None

        # a list of nodes that makes use of the symbol
        self._references = []

    def _resolve(self, context):
        raise NotImplementedError

    def _store(self, context, value):
        """ Storing the value of the variable to the context """
        # Not doing anything to the context by default
        pass

    def _add_reference(self, node):
        self._references.append(weakref.ref(self))
        ref_id = len(self._references)
        return Ref(self, node, ref_id)

    def _has_reference(self):
        return len(self._references) > 0

    def eval(self, function):
        """ Evaluates an expression or a function on the symbol. see vmad.core.stdlib.eval. """
        return stdlib.eval(self, function)

    # workaround numpy's operator overrides, which
    # is higher priority than rxxx methodds (radd) defined here.
    # this will be deprecated and we will need to add an
    # __array_ufunc__ to dispatch this in the future.
    __array_priority__ = 15.0

    def __add__(self, other): return stdlib.add(self, other, __stacklevel__=-3)
    def __radd__(self, other): return stdlib.add(other, self, __stacklevel__=-3)
    def __sub__(self, other): return stdlib.sub(self, other, __stacklevel__=-3)
    def __rsub__(self, other): return stdlib.sub(other, self, __stacklevel__=-3)
    def __mul__(self, other): return stdlib.mul(self, other, __stacklevel__=-3)
    def __rmul__(self, other): return stdlib.mul(other, self, __stacklevel__=-3)
    def __mod__(self, other): return stdlib.mod(self, other, __stacklevel__=-3)
    def __rmod__(self, other): return stdlib.mod(other, self, __stacklevel__=-3)
    def __truediv__(self, other): return stdlib.div(self, other, __stacklevel__=-3)
    def __rtruediv__(self, other): return stdlib.div(other, self, __stacklevel__=-3)
    def __pow__(self, other, modulo=None):
        if modulo is not None:
            raise ValueError("pow with modulo is not supported")
        return stdlib.pow(self, other, __stacklevel__=-3)
    def __rpow__(self, other):
        raise
        return stdlib.pow(other, self, __stacklevel__=-3)

    def __floordiv__(self, other): raise TypeError("floor div is not supported in autodiff")
    def __rfloordiv__(self, other): raise TypeError("floor div is not supported in autodiff")

    def __neg__(self): return stdlib.neg(self, __stacklevel__=-3)
    def __pos__(self): return stdlib.pos(self, __stacklevel__=-3)
    def __abs__(self): return stdlib.abs(self, __stacklevel__=-3)

class Ref(object):
    """
        A reference is the relation between a symbol and a node.
    """
    def __init__(self, symbol, node, ref_id):
        self.symbol = symbol
        # the node variable is currently not used anywhere
    #    self.node = node
        self.ref_id = ref_id

    def __repr__(self):
        return "&[%s:%d]" % (repr(self.symbol), self.ref_id)

    def get_symbol_names(self):
        """ Returns a list of symbol names that are referenced by
            this object, recursively
        """
        from .symbol import Symbol
        l = set([])
        symbol = self.symbol
        # recusrively get the name of the parents
        # to handle AttrSymbol and CallSymbol
        while hasattr(symbol, '_parent'):
            if isinstance(symbol, Symbol):
                l = l.union([symbol._name])
            symbol = symbol._parent
        return l

class Symbol(BaseSymbol):
    """ A symbol remembers the model only for eaiser model construction.

        A symbol is bound to a model; this is useful
        for the operators to infer the model during model construction.
    """
    def __init__(self, name, model=None):
        assert isinstance(name, str)

        BaseSymbol.__init__(self)

        self._name = name


        self._model_ref = None

        if model is not None:
            # anchor it now
            model.anchor(self)

    @property
    def _model(self):
        from .model import ModelInTransient
        if self._model_ref is None:
            return None
        if isinstance(self._model_ref, ModelInTransient):
            return self._model_ref
        else:
            return self._model_ref()

    @property
    def _anchored(self):
        return self._model_ref is not None

    @property
    def _transient(self):
        from .model import ModelInTransient
        return isinstance(self._model_ref, ModelInTransient)

    def __repr__(self):
        return self._name

    @property
    def _vjp_name(self):
        return '_' + self._name

    @property
    def _jvp_name(self):
        return self._name + '_'

    def _resolve(self, context):
        if self._name not in context:
            raise ResolveError("Symbol %s does not exist in the context" % self._name)
        return context[self._name]

    def _store(self, context, value):
        context[self._name] = value

class List(BaseSymbol):
    def __init__(self, value):
        BaseSymbol.__init__(self)
        self._items = value

    def __repr__(self):
        return "L%s" % (str(self._items))

    def _resolve(self, context):
        return [v._resolve(context) for v in self]

    def __len__(self):
        return len(self._items)

    def __getitem__(self, i):
        return self._items[i]

    def __iter__(self):
        for v in self._items:
            yield v

    def __reversed__(self):
        for v in reversed(self._items):
            yield v

    def _store(self, context, value):
        for var, v in zip(self, value):
            var._store(context, v)

    def _add_reference(self, node):
        self._references.append(weakref.ref(self))
        ref_id = len(self._references)
        return ListRef(self, node, ref_id)

    def _has_reference(self):
        return (
            BaseSymbol._has_reference(self) or
            any(v._has_reference() for v in self)
        )

class ListRef(Ref):
    def __init__(self, symbol, node, ref_id):
        Ref.__init__(self, symbol, node, ref_id)
        self.symbol = symbol
        # make sure the last reference shows up first.
        self.items = list(reversed([v._add_reference(node) for v in reversed(symbol)]))

    def __repr__(self):
        return "& %s" % (str(self.items))

    def __iter__(self):
        return iter(self.items)

    def get_symbol_names(self):
        r = set()
        for ref in self.items:
            r = r.union(ref.get_symbol_names())
        return r

class IgnoreSymbol(BaseSymbol):
    """ A IgnoreSymbol is a placeholder for Literals in the autodiff
        gradient propagation.
    """
    def __init__(self):
        BaseSymbol.__init__(self)

    def __repr__(self):
        return "IGN"

    # Never referenced; allow bypassing unused nodes.
    def _has_reference(self):
        return False

IgnoredGradient = IgnoreSymbol()

class Literal(BaseSymbol):
    """ A literal is a special symbol that does not resolve with a context.

        Literals do not participate in gradient propagation.
    """
    def __init__(self, value):
        BaseSymbol.__init__(self)
        self._value = value

    def __repr__(self):
        return "L(%s)" % (str(self._value))

    def _resolve(self, context):
        return self._value

    def __getattr__(self, attrname):
        if attrname.startswith('_'):
            raise AttributeError
        return AttrSymbol(self, attrname)

class AttrSymbol(Literal):
    """ Represents accessing a member attribute of a symbol.

        The attribute is always treated as a literal. The suitable context
        is saying getting the size of an array.
    """
    def __init__(self, parent, attrname):
        Literal.__init__(self, None)
        self._parent = parent
        self._attrname = attrname

    def _resolve(self, context):
        return getattr(self._parent._resolve(context), self._attrname)

    def __repr__(self):
        return "%s.%s" % (str(self._parent), self._attrname)

class CallSymbol(Literal):
    """ Represents calling a member attribute of a symbol.
    """
    def __init__(self, parent, args, kwargs):
        Literal.__init__(self, None)
        self._parent = parent
        self._args = args
        self._kwargs = kwargs

    def _resolve(self, context):
        return self._parent._resolve(context)(*self._args, **self._kwargs)

class GetItemSymbol(Literal):
    """ Represents getting an item of a symbol.
    """
    def __init__(self, parent, index):
        Literal.__init__(self, None)
        self._parent = parent
        self._index = index

    def _resolve(self, context):
        return self._parent._resolve(context)[index]

class ZeroLiteral(Literal):
    """ A ZeroLiteral is specially used to mark zeros in gradient propagation

    """
    def __init__(self):
        Literal.__init__(self, None)

    def __repr__(self):
        return "[ZERO]"

    def _resolve(self, context):
        return 0

def assymbol(obj):
    """ Make a symbol out of an input Python object.

    """
    # cast list or tuple to a list object:
    if type(obj) in (list, tuple):
        obj = List([assymbol(i) for i in obj])

    # not a Symbol? Must be some raw python object
    # intended to be used as a Literal
    if not isinstance(obj, BaseSymbol):
        obj = Literal(obj)

    return obj



from vmad.core.symbol import Symbol
from vmad.core.operator import operator
from vmad import Builder

def test_symbol_eval():
    with Builder() as m:
        a = m.input('a')
        m.output(
                b=a.eval(lambda a: len(a)),
                c=a.eval("len(x)"),
                )

    m.compute(['b', 'c'], init=dict(a=[1, 2, 3]))
    m.compute_with_vjp(init=dict(a=[1, 2, 3]), v=dict(_b=1.0, _c=1.0))
    m.compute_with_jvp(['b', 'c'], init=dict(a=[1, 2, 3]), v=dict(a_=1.0))

def test_symbol_add():
    with Builder() as m:
        a, b = m.input('a', 'b')
        m.output(
                c = (a + b) * b / b ** 2
                )

    c = m.compute('c', init=dict(a=2, b=3))
    assert round(c, 4) == round((2 + 3) * 3 / 3 ** 2, 4)

def test_subclass_symbol():

    class MySymbol(Symbol):
        @operator
        class __add__:
            ain = 'x', 'y'
            aout = 'z'
            def apl(self, x, y):
                return dict(z=x + y)
            def vjp(self, _z):
                return dict(_x=_z, _y=_z)
            def jvp(self, x_, y_):
                return dict(z_ = x_+ y_)

    with Builder() as m:
        a, b = m.input(MySymbol('a'), MySymbol('b'))
        c = a + b
        m.output(c=c)

    m.compute('c', init=dict(a=1, b=2))
    m.compute_with_vjp(init=dict(a=1, b=2), v=dict(_c=1.0))
    m.compute_with_jvp(['c'], init=dict(a=1, b=2), v=dict(a_=1.0, b_=1.0))




from . import autodiff

class Record(object):
    """ A record on the tape. 

        A record contains the node and the resolved arg symbols to the node.

    """
    def __init__(self, node, impl_kwargs):
        self.node = node
        self.impl_kwargs = impl_kwargs

    def __repr__(self):
        return '%s / %s' % (self.node, self.impl_kwargs)

class Tape(list):
    def __init__(self, model, init):
        self.model = model
        self.init = init
        self._completed = False

    def finalize(self, out):
        """ Finalize the tape, with a set of computed outputs.

            Parameters
            ----------
            out : dict / OrderedDict
        """
        assert isinstance(out, dict)
        self.out = out.copy()
        self._completed = True

    def append(self, node, impl_kwargs):
        assert not self._completed

        list.append(self, Record(node, impl_kwargs))

    def get_vjp_vout(self):
        return ['_' + varname for varname in self.init.keys()]

    def get_jvp_vout(self):
        return [varname + '_' for varname in self.out.keys()]

    def get_vjp(self):
        assert self._completed
        return autodiff.vjpmodel(self)

    def get_jvp(self):
        assert self._completed
        return autodiff.jvpmodel(self)

    def compute_jvjp(self, vout, aout, init):
        jvp = self.get_jvp()
        aout_ = [a + '_' for a in aout]
        t = jvp.compute(aout_, init)
        vjp = self.get_vjp()
        p = vjp.compute(vout, init=dict([('_' + a, t1) for a, t1 in zip(aout, t)]))
        return p




from __future__ import print_function
from pprint import pprint
from vmad.lib import linalg
from vmad.core import stdlib
import numpy

from vmad import Builder
from vmad.testing import BaseVectorTest

class Test_pack_complex(BaseVectorTest):
    x = numpy.arange(10) # will pack to complex of x + x * 1j
    y = numpy.sum(2 * x ** 2)

    def model(self, x):
        # FIXME: pack_complex only works with to_scalar
        # but it doesn't work with mul and conj
        # there is a missing negative sign. one of these is wrong.
        c = linalg.pack_complex(x, x)
        c = linalg.to_scalar(c)
        return c

class Test_conj_mul(BaseVectorTest):
    x = numpy.arange(10) # will pack to complex of x + x * 1j
    y = x ** 2

    def model(self, x):
        # FIXME: pack_complex only works with to_scalar
        # but it doesn't work with mul and conj
        # there is a missing negative sign. one of these is wrong.
        x = x * 1j
        c = linalg.conj(x)
        return x * c

class Test_unpack_complex(BaseVectorTest):
    x = numpy.arange(10) # will pack to complex of x + x * 1j
    y = x * 2

    def model(self, x):
        c = linalg.pack_complex(x, x)
        r, i = linalg.unpack_complex(c)
        return linalg.add(r, i)

class Test_reshape(BaseVectorTest):
    x = numpy.arange(10) 
    y = x.reshape(5, 2)

    def model(self, x):
        c = linalg.reshape(x, (5, 2))
        return c

class Test_einsum(BaseVectorTest):
    x = numpy.arange(10)
    y = numpy.sum(x ** 2).reshape(1)

    def model(self, x):
        # testing uncontracted dimensions
        a = linalg.reshape(x, (10, 1, 1))
        b = linalg.reshape(x, (1, 10))
        return linalg.einsum("abc, ca->b", [a, b])

class Test_einsum2(BaseVectorTest):
    x = numpy.arange(10)
    y = numpy.sum(x ** 2)

    def model(self, x):
        # testing contraction
        a = linalg.reshape(x, (10,))
        b = linalg.reshape(x, (10))
        return linalg.einsum("i, i->", [a, b])

class Test_mul1(BaseVectorTest):
    x = numpy.arange(10)
    y = x

    def model(self, x):
        return linalg.mul(x, 1.0)

class Test_mul2(BaseVectorTest):
    x = numpy.arange(10) + 1
    y = x
    def model(self, x):
        y1 = linalg.pow(x, 0.3)
        y2 = linalg.pow(x, 0.7)
        return linalg.mul(y1, y2)

class Test_mul3(BaseVectorTest):
    x = numpy.arange(10) 
    y = x

    def model(self, x):
        return linalg.mul(1, x)

class Test_div2(BaseVectorTest):
    x = numpy.arange(1, 10)
    y = x

    def model(self, x):
        y1 = linalg.pow(x, 1.7)
        y2 = linalg.pow(x, 0.7)
        return linalg.div(y1, y2)

class Test_mod(BaseVectorTest):
    x = numpy.arange(1, 4) * 2 - 3
    y = x % (x * 0.1 + 1.0)

    def model(self, x):
        return linalg.mod(x, x * 0.1 + 1.0)

class Test_sum(BaseVectorTest):
    x = numpy.arange(10)
    y = x.reshape(5, 2).sum(axis=0)

    def model(self, x):
        x = linalg.reshape(x, (5, 2))

        return linalg.sum(x, axis=0)

class Test_pow(BaseVectorTest):
    x = numpy.arange(10)
    y = x ** 2.0

    def model(self, x):
        return linalg.pow(x, 2.0)

class Test_abs(BaseVectorTest):
    x = numpy.arange(10) - 4.5 # avoid 0 because numerial is bad.
    y = (abs(x) + x)

    def model(self, x):
        return linalg.add(linalg.abs(x), x)

class Test_log(BaseVectorTest):
    logx = 1 + numpy.arange(10)
    x = numpy.exp(logx)
    y = logx
    epsilon = 1e-3 # enlarge epsilon for loss of precision in log.
    def model(self, x):
        return linalg.log(x)

class Test_add(BaseVectorTest):
    x = numpy.arange(10)
    y = (x + 5.0)

    def model(self, x):
        return linalg.add(x, 5.0)

class Test_copy(BaseVectorTest):
    x = numpy.arange(10)
    y = x

    def model(self, x):
        return linalg.copy(x)

class Test_stack(BaseVectorTest):
    x = numpy.arange(10)
    y = numpy.stack([x, x])

    def model(self, x):
        return linalg.stack([x, x], axis=0)

class Test_take(BaseVectorTest):

    x = numpy.arange(10)
    y = numpy.array(4)

    def model(self, x):
        return linalg.sum(linalg.take(x, [2, 2], axis=0), axis=0)

from vmad.core.stdlib import watchpoint
class Test_take_vector(BaseVectorTest):

    x = numpy.arange(10)
    y = numpy.array(4)
    def model(self, x):
        a = linalg.take(x, [[2, 2], [2, 2]], axis=0)
        b = linalg.take(a, 0, axis=0)
        c = linalg.take(b, 0, axis=0)
        #d = linalg.take(b, 0, axis=0)
        #watchpoint(b, lambda b:print('b=', b))
        #watchpoint(c, lambda c:print('c=', c))
    #    watchpoint(d, lambda d:print('d=', d))
        return c + c

class Test_concatenate(BaseVectorTest):

    x = numpy.arange(10).reshape(5, 2)
    y = numpy.concatenate([x, x], axis=1)

    def model(self, x):
        #a = linalg.concatenate([x, x], axis=0)
        b = linalg.concatenate([x, x], axis=1)
        return b

class Test_transpose(BaseVectorTest):

    x = numpy.arange(10).reshape(5, 2)
    y = numpy.transpose(x, (1, 0))

    def model(self, x):
        #a = linalg.concatenate([x, x], axis=0)
        b = linalg.transpose(x, (1, 0))
        return b

class Test_sumat(BaseVectorTest):
    x = numpy.arange(10).reshape(5, 2)
    at = [0, 1, 3]
    y = numpy.add.reduceat(x, at, axis=0)

    def model(self, x):
        return linalg.sumat(x, at=self.at, axis=0)

class Test_broadcast(BaseVectorTest):
    x = numpy.arange(10).reshape(5, 1, 2)
    shape = [3, 5, 2, 2]
    y = numpy.broadcast_to(x, shape)

    def model(self, x):
        return linalg.broadcast_to(x, self.shape)

def test_take_chained():
    from vmad import autooperator
    from numpy.testing import assert_array_equal

    @autooperator('x->y')
    def func(x):
        b = 0
        for i in range(3):
            b = b + linalg.take(x, i, axis=0)
        return b

    (y,), (_x, ) = (func.build().compute_with_vjp(init=dict(x=numpy.array([3, 4, 5])), v=dict(_y=1.0)))
    assert y == 12
    assert_array_equal(_x, (1, 1, 1))

def test_variance_vjp():
    # test case contributed by BiweiDai
    # not necessarily the best way to compute variance.
    from vmad import autooperator
    from numpy.testing import assert_array_equal
    @autooperator('x->y')
    def test(x):
        size = stdlib.eval(x, lambda x: numpy.prod(x.shape))
        mean = linalg.sum(x, axis=None) / size
        mean = linalg.broadcast_to(mean, stdlib.eval(x, lambda x : x.shape))
        x = x - mean
        y = linalg.sum(x**2, axis=None) / size
        return y

    model = test.build()
    y, [vjp] = model.compute_with_vjp(init=dict(x=numpy.array([1, 3])), v=dict(_y=1.0))
    assert_array_equal(vjp, (-1, 1))
    assert_array_equal(y, 1)



import inspect
from .operator import _make_primitive, Operator, unbound, _to_ordereddict
from .model import Builder
from .context import Context
from .error import ModelError

class AutoOperator(Operator):
    """ Base class to support operators with on demand tape and prerecorded tape """

    def __init__(self, prototype, argnames):
        Operator.__init__(self, prototype)

        impl = unbound(prototype.main)

        # use the argnames of main function
        if argnames is None:
            argnames = impl.__code__.co_varnames[1:impl.__code__.co_argcount]

        self.argnames = argnames
        self.hyperargs = {}
        self.__bound_tape__ = None
        self.__bound_model__ = None
        self.__tapename__ = '$$tape$$'

    @property
    def forgetful(self):
        obj = AutoOperator(self.prototype, self.argnames)
        obj.__bound_tape__ = self.__bound_tape__
        obj.__bound_model__ = self.__bound_model__
        obj.__tapename__ = None

        return obj

    @property
    def apl(self):
        return self.get_apl(tapename=self.__tapename__)

    def get_apl(self, tapename):
        """ get an apl primitive, if tapename is not None, the primitive returns the tape as well. """
        def apl(node, **kwargs):
            # node.operator shall be the same as self.
            y = node.operator._compute(kwargs, tapename=tapename)
            return y

        def rcd(node, **kwargs):
            return kwargs

        outnames = list(self.aout.keys())

        return _make_primitive(self, 'apl', apl, argnames=self.argnames, outnames=outnames, record_impl=rcd)

    @property
    def vjp(self):
        return self.get_vjp(tapename=self.__tapename__)

    def get_vjp(self, tapename):
        # add v arguments
        argnames_vjp = list(self.argnames)
        if tapename is not None:
            argnames_vjp.append(tapename)

        for argname in self.aout:
            argnames_vjp.append('_' + argname)

        def vjp(node, **kwargs):
            if tapename is not None:
                tape = kwargs.pop(tapename)
            else:
                # run the apl operator to obtain the tape
                y = node.operator._compute(kwargs, '$$tape$$')
                tape = y['$$tape$$']

            v =    [(a, kwargs[a]) for a in node.primitive.ain.keys() if a.startswith('_')]

            vjp = tape.get_vjp()
            vjpvout = tape.get_vjp_vout()
            vjpout = vjp.compute(vjpvout, init=v)
            return dict(zip(vjpvout, vjpout))

        return _make_primitive(self, 'vjp', vjp, argnames=argnames_vjp)

    @property
    def jvp(self):
        return self.get_jvp(tapename=self.__tapename__)

    def get_jvp(self, tapename):
        argnames_jvp = list(self.argnames)
        if tapename is not None:
            argnames_jvp.append(tapename)

        for argname in self.ain:
            argnames_jvp.append(argname + '_')

        def jvp(node, **kwargs):
            if tapename is not None:
                tape = kwargs[tapename]
            else:
                # run the apl operator to obtain the tape
                y = node.operator._compute(kwargs, '$$tape$$')
                tape = y['$$tape$$']

            v =    [(a, kwargs[a]) for a in node.primitive.ain.keys() if a .endswith('_')]

            jvp = tape.get_jvp()
            jvpvout = tape.get_jvp_vout()
            jvpout = jvp.compute(jvpvout, init=v)
            return dict(zip(jvpvout, jvpout))

        return _make_primitive(self, 'jvp', jvp, argnames=argnames_jvp)

    def _compute(self, kwargs, tapename):
        if self.__bound_tape__ is not None:
            tape = self.__bound_tape__
            y = {}
            y.update(tape.out)
        else :
            m = self._build(kwargs)
            init = [(a, kwargs[a]) for a in self.ain.keys()]

            vout = list(self.aout.keys())
            if tapename is not None:
                y, tape = m.compute(vout, init=init, return_dict=True, return_tape=True)
            else:
                y = m.compute(vout, init=init, return_dict=True, return_tape=False)

        if tapename is not None:
            y[tapename] = tape

        return y

    def _build(self, kwargs):
        if self.__bound_model__ is not None:
            m = self.__bound_model__
            return m

        impl = unbound(self.prototype.main)

        model_args = {}
        # copy extra args of the main function
        for argname in self.apl.argnames:
            if argname not in self.ain:
                model_args[argname] = kwargs[argname]

        with Builder() as m:
            # add input args as variables
            for argname in self.ain:
                model_args[argname] = m.input(argname)
            r = impl(m, **model_args)
            # assert outputs are generated
            for argname in self.aout:
                if argname not in r:
                    raise ModelError("output arg '%s' is not produced by the model" % argname)
            m.output(**r)
        return m

    # FIXME: add docstring / argnames
    # shall be the list of extra args
    def build(__instance__1234__, **kwargs):
        """ Create a computing graph model for the operator with the given hyper parameters """
        self = __instance__1234__
        for argname in kwargs:
            if argname in self.ain:
                raise ModelError("argname %s is an input, shall not be used to produce a model" % argname)

        return self._build(kwargs)


    def precompute(self, **kwargs):
        """ Create a bound autooperator where the hyperparameter and parameters are both given; the model
            is compuated, and the tape is recorded.

            Instantiating the returned operator will be an exact replay of the tape, regardless of the parameters
        """
        y = self._compute(kwargs, '$$tape$$')
        m = self._build(kwargs)
        tape = y['$$tape$$']

        # create a new operator, because we need new primitives that points to this operator.
        obj = AutoOperator(self.prototype, argnames=self.argnames)
        obj.__bound_tape__ = tape
        obj.__bound_model__ = m

        return obj

    def bind(self, **hyperargs):
        """ Create a bound autooperator where the hyperparameter are already given.

            Instantiating the returned operator no longer requires the hyperparameters.
        """
        for argname in hyperargs:
            if argname in self.ain:
                raise ModelError("argname %s is an input, shall not be used to produce a model" % argname)

        m = self._build(hyperargs)

        # remove those args already defined
        argnames = tuple([
                argname for argname in self.apl.argnames if argname not in hyperargs
                ])

        # create a new operator, because we need new primitives that points to this operator.
        obj = AutoOperator(self.prototype, argnames=argnames)
        obj.__bound_model__ = m
        obj.hyperargs = hyperargs

        return obj

def _autograd(func, ain, aout):

    def main(__unused_argument_3333__, *args, **kwargs):
        kwargs = kwargs.copy()
        # merge the positional arguments
        kwargs.update(dict(zip(ain, args)))

        # always call with the kwargs syntax, such that
        r = func(**kwargs)
        # normalize the output to a dict.
        if isinstance(r, tuple):
            r = dict(zip(aout, r))
        elif not isinstance(r, dict) and len(aout) == 1:
            r = dict(zip(aout, [r]))

        return r

    argnames = func.__code__.co_varnames[:func.__code__.co_argcount]
    prototype = type(func.__name__, (),
            dict(
                ain=ain,
                aout=aout,
                main=main,
                )
            )
    return AutoOperator(prototype, argnames=argnames)

def _parse_annotations(func):
    annotations = getattr(func, '__annotations__', {})
    if len(annotations) == 0:
        raise ValueError("Must provide ain, aout or function annotations. see docstring of autograd")
    ain = [
        (key, value)
        for key, value in annotations.items() if key != 'return'
    ]
    aout = annotations['return']

    return ain, aout, func

def _parse_autograd_spec(args):
    if len(args) == 1:
        if hasattr(args[0], '__call__'):
            return _parse_annotations(args[0])

        spec = args[0]
        split = spec.index('->')
        left = spec[:split]
        right = spec[split+2:]

        ain = [a.strip() for a in left.split(',')]
        aout = [a.strip() for a in right.split(',')]
    else:
        split = args.index('->')
        ain = args[:split]
        aout = args[split + 1:]
    return ain, aout, None

def autooperator(*args):
    """ autooperator creates a operator from a prototype class
        or a function that uses vmad to build a model.

        The function or the main method must only use vmad operators
        on the input autooperator symbols.

        1. Use Python 3 annotation syntax to declare
        the input and output arguments.

        .. code ::

            @autooperator
            def function(x : '*', y : '*', n) -> 'z':
                ...
                return z

        2. Explicitly provide the lis of input and output
           arguments.

        .. code ::

            @autooperator('x', 'y', '->', 'z')
            def function(x, y, n):
                ...
                return z

        or 

        .. code ::

            @autooperator('x, y->z')
            def function(x, y, n):
                ...
                return z

        or
        .. code ::

            @autooperator('x, y->z')
            def function(n, **kwargs):
                ...
                return z

        3. Use a class
         Create an operator with automated differentiation.

        ain : input arguments
        aout : output arguments

        main : function(model, ...) building the model;
                the arguments are the input symbols
                returns the dict of output symbols,
                shall match the output arguments

        see the example below in this file.

        .. code ::
            @autooperator
            class function:
                ain = 'x', 'y'
                aout = 'z'
                def main(model, x, y, n):
                    ...
                    return dict(z = ....)

    """
    if inspect.isclass(args[0]):
        return AutoOperator(args[0], argnames=None)

    ain, aout, func = _parse_autograd_spec(args)
    def wrapped(func):
        return _autograd(func, ain, aout)

    if func is None:
        return wrapped
    else:
        return wrapped(func)




from vmad import Builder
from numpy.testing import assert_array_equal, assert_allclose

import numpy

class BaseScalarTest:
    """ Basic correctness of gradient against numerical with to_scalar """

    to_scalar = None # operator for norm-2 scalar

    x = numpy.arange(10)  # free variable x
    x_ = numpy.eye(10)    # a list of directions of x_ to check for directional gradients.

    y = sum(x ** 2)       # expected output variable y, scalar
                          # NotImplemented to bypass the value comparison
    epsilon = 1e-3
    rtol = 1e-6
    atol = 1e-10
    def inner(self, x, y):
        return numpy.sum(x * y)

    def model(self, x):
        return x          # override and build the model will be converted to a scalar later.

    def setup(self):
        with Builder() as m:
            x = m.input('x')
            x = self.model(x)
            # avoid using the bound function.
            y = type(self).to_scalar(x)
            m.output(y=y)

        self.m = m

        y_ = []
        for x_ in self.x_:
            # run a step along x_
            xl = self.x - x_ * (self.epsilon * 0.5)
            xr = self.x + x_ * (self.epsilon * 0.5)

            # numerical
            yl = self.m.compute(init=dict(x=xl), vout='y', return_tape=False)
            yr = self.m.compute(init=dict(x=xr), vout='y', return_tape=False)

            y_1 = (yr - yl) / self.epsilon

            y_.append(y_1)

        y, tape = self.m.compute(init=dict(x=self.x), vout='y', return_tape=True)
        self.tape = tape
        self.y_ = y_

    def test_opr(self):
        init = dict(x=self.x)
        y1 = self.m.compute(vout='y', init=init, return_tape=False)

        if self.y is not NotImplemented:
            # correctness
            assert_allclose(y1, self.y, rtol=self.rtol, atol=self.atol)

    def test_powerfulness(self):
        import numpy

        if numpy.allclose(self.y_, 0):
            raise AssertionError("The test case is not powerful enough, since all derivatives at this point are zeros")

    def test_jvp_finite(self):
        jvp = self.tape.get_jvp()

        for x_, y_ in zip(self.x_, self.y_):
            init = dict(x_=x_)
            y_1 = jvp.compute(init=init, vout='y_', return_tape=False)

            assert_allclose(y_1, y_, rtol=self.rtol, atol=self.atol)

    def test_vjp_finite(self):
        vjp = self.tape.get_vjp()

        init = dict(_y=1.0)
        _x = vjp.compute(init=init, vout='_x', return_tape=False)

        for x_, y_ in zip(self.x_, self.y_):
            assert_allclose(self.inner(_x, x_), y_, rtol=self.rtol, atol=self.atol)

class BaseVectorTest:
    """ Basic correctness of gradient against numerical for vector functions """

    x = numpy.arange(10)  # free variable x
    x_bases = None    # a list of directions of x_ to check for directional gradients.
                               # if None, generate based on the shape of x.
    y = x ** 2 # expected output variable y

    y_bases = None    # a list of directions of y_ to check for directional gradients.
                               # if None, generate based on the shape of y.


    epsilon = 1e-7

    rtol = 1e-6
    atol = 1e-10
    def allclose(self, x, y):  # measuring the 
        assert numpy.shape(x) == numpy.shape(y)
        return numpy.allclose(x, y, rtol=self.rtol, atol=self.atol)

    def inner(self, a, b):
        return numpy.sum(a * b)

    def model(self, x):
        return x          # override and build the model will be converted to a scalar later.

    def _make_bases(self, x):
        x_ = []
        eye = numpy.eye(x.size)
        for row in eye:
            x_.append(row.reshape(x.shape).copy())
        return x_

    def _make_model(self):
        with Builder() as m:
            x = m.input('x')
            y = self.model(x)
            m.output(y=y)
        return m

    def _make_finite_jacobian(self):
        y_ = []
        for x_ in self.x_bases:
            # run a step along x_
            xl = self.x - x_ * (self.epsilon * 0.5)
            xr = self.x + x_ * (self.epsilon * 0.5)

            # numerical
            yl = self.m.compute(init=dict(x=xl), vout='y', return_tape=False)
            yr = self.m.compute(init=dict(x=xr), vout='y', return_tape=False)

            y_1 = (yr - yl) / self.epsilon

            y_.append(y_1)

        return y_

    def setup(self):

        if hasattr(self.y, '__call__'):
            self.y = self.y(self.x)

        self.m = self._make_model()

        if self.x_bases is None:
            self.x_bases = self._make_bases(self.x)

        if self.y_bases is None:
            self.y_bases = self._make_bases(self.y)

        self.finite_jvp = self._make_finite_jacobian()

        if numpy.allclose(self.finite_jvp, 0):
            raise AssertionError("The test case is not powerful enough, since all derivatives at this point are zeros")

        y, tape = self.m.compute(init=dict(x=self.x), vout='y', return_tape=True)

        self.tape = tape

        self._evaluated_y = y


    def test_opr(self):
        # if this is wrong, apl is wrong
        assert_allclose(self._evaluated_y, self.y, rtol=self.rtol, atol=self.atol)

    def test_jvp_finite(self):
        jvp = self.tape.get_jvp()

        for x_, y_ in zip(self.x_bases, self.finite_jvp):
            init = dict(x_=x_)
            y_1 = jvp.compute(init=init, vout='y_', return_tape=False)

            if not self.allclose(y_1, y_):
                raise AssertionError("jvp comparison failed")

    def test_vjp_finite(self):
        vjp = self.tape.get_vjp()

        for _y in self.y_bases:
            init = dict(_y=_y)
            _x = vjp.compute(init=init, vout='_x', return_tape=False)

            for x_, y_ in zip(self.x_bases, self.finite_jvp):
                v1 = self.inner(_y, y_)
                v2 = self.inner(_x, x_)
                if not self.allclose(v1, v2):
                    raise AssertionError("jvp comparison failed")




from vmad.core.model import Builder
from vmad.core.stdlib import add
import pytest

def test_error_infer():
    from vmad.core.error import InferError
    with Builder() as m:
        with Builder() as m2:
            a = m.input('a')
            b = m2.input('b')
            with pytest.raises(InferError):
                c = add(x1=a, x2=b)

def test_error_bad_arg1():
    from vmad.core.error import BadArgument
    with Builder() as m:
        a = m.input('a')
        with pytest.raises(BadArgument):
            add(1, 1, 1)

def test_error_bad_arg2():
    from vmad.core.error import BadArgument
    with Builder() as m:
        a = m.input('a')
        with pytest.raises(BadArgument):
            add(1, x1=1, x2=2)

def test_error_missing():
    from vmad.core.error import MissingArgument
    with Builder() as m:
        a = m.input('a')
        with pytest.raises(MissingArgument):
            add(x2=1)

def test_error_overwrite():
    from vmad.core.error import OverwritePrecaution
    with Builder() as m:
        a = m.input('a')
        with pytest.raises(OverwritePrecaution):
            add(x1=a, x2=a, y=a)

def test_error_many_output():
    from vmad.core.error import DuplicatedOutput
    with Builder() as m:
        a = m.input('a')
        with pytest.raises(DuplicatedOutput):
            m.output(a=a)
            m.output(a=a)


def test_error_unexpected_output():
    from vmad.core.error import UnexpectedOutput
    with Builder() as m:
        a = m.input('a')
        m.output(a=a)

    with pytest.raises(UnexpectredOutput):
        m.compute(vout='b', init=dict(a=1.0))

def test_error_unexpected_output():
    from vmad.core.error import ResolveError
    with Builder() as m:
        a = m.input('a')
        m.output(a=a)

    with pytest.raises(ResolveError):
        m.compute(vout='a', init={})



from vmad.core.operator import operator

# special operator for marking an output
@operator
class terminal:
    ain  = 'x'
    aout = 'y'

    def apl(node, x):
        return dict(y=x)
    def vjp(node, _y):
        return dict(_x=_y)
    def jvp(node, x_):
        return dict(y_=x_)


from .utils import eval, watchpoint
from .utils import assert_isinstance, assert_true

from .operators import pos, neg, add, sub, mul, div, pow, abs, mod




from __future__ import print_function
from pprint import pprint
from vmad.lib import linalg, mpi

import numpy

from vmad.testing import BaseScalarTest
from mpi4py import MPI

class Test_allreduce(BaseScalarTest):
    to_scalar = staticmethod(linalg.to_scalar)

    comm = MPI.COMM_WORLD

    x = comm.rank + 1.0
    y = comm.allreduce(x) ** 2

    x_ = numpy.eye(1)

    # self.x is distributed, thus allreduce along the rank axis.
    def inner(self, a, b):
        return self.comm.allreduce(numpy.sum(a * b))

    def model(self, x):
        return mpi.allreduce(x, self.comm)


class Test_allbcast(BaseScalarTest):
    to_scalar = staticmethod(lambda x: x)
    comm = MPI.COMM_WORLD
    x = 2.0
    y = comm.allreduce(x * (comm.rank + 1))

    x_ = numpy.eye(1)

    # self.x is universal, thus no special allreduce here.
    def inner(self, a, b):
        return numpy.sum(a*b)

    def model(self, x):
        x = mpi.allbcast(x, self.comm)
        x = x * (self.comm.rank + 1)
        return mpi.allreduce(x, comm=self.comm)



"""
    Routines to define an operator

    use @operator decorator on a class to define an operator.

    Example: see the source code of :class:`add`

    use @nested.modeloperator to define a nested operator, where
    you only need to define a model and the ain, aout

"""
from collections import OrderedDict

class ZeroGradientType(int):
    def __new__(self):
        return int.__new__(ZeroGradientType, 0)

    # for zip
    def __iter__(self):
        while True:
            yield ZeroGradient

    def __array__(self, dtype):
        import numpy
        return numpy.array(0, dtype=dtype)

    @staticmethod
    def is_zero(obj):
        return isinstance(obj, ZeroGradientType) or obj is 0 or obj is 0.0

# A universal zero used in jvp and vjp
ZeroGradient = ZeroGradientType()

class EmptyPrimitive:
    argnames = []

class Operator(object):
    def __init__(self, prototype):
        self.prototype = prototype
        self.ain = _to_ordereddict(prototype.ain)
        self.aout = _to_ordereddict(prototype.aout)

        # alias all static methods from the prototype
        for name, value in prototype.__dict__.items():
            if isinstance(value, staticmethod):
                setattr(self, name, getattr(prototype, name))

    def __call__(self, *args, **kwargs):
        stacklevel = kwargs.pop('__stacklevel__', -2)
        kwargs, kwout = self.apl._parse_args(args, kwargs)
        vout = self.apl.create_node(kwargs, kwout, stacklevel=stacklevel)

        if len(self.aout) == 1:
            return next(iter(vout.values()))
        else:
            return list(vout.values())

    def __get__(self, instance, owner):
        if instance is not None:
            return InstanceOperator(self, instance)
        else:
            return self

    @property
    def apl(self):
        if hasattr(self.prototype, 'rcd'):
            record_impl = unbound(self.prototype.rcd)
        else:
            record_impl = record_copy_autodiff

        return _make_primitive(self, 'apl', unbound(self.prototype.apl),
                record_impl=record_impl)

    @property
    def vjp(self):
        if hasattr(self.prototype, 'vjp'):
            return _make_primitive(self, 'vjp', unbound(self.prototype.vjp))
        else:
            return EmptyPrimitive

    @property
    def jvp(self):
        if hasattr(self.prototype, 'jvp'):
            return _make_primitive(self, 'jvp', unbound(self.prototype.jvp))
        else:
            #return EmptyPrimitive
            argnames_jvp = []

            for argname in self.apl.argnames:
                if argname in self.ain:
                    argnames_jvp.append(argname + '_')
                else:
                    argnames_jvp.append(argname)

            return _make_primitive(self, 'jvp', default_jvp, argnames=argnames_jvp)

class InstanceOperator(Operator):
    def __init__(self, base, instance):
        self.base = base
        self.instance = instance

    def __call__(self, *args, **kwargs):
        return self.base(self.instance, *args, **kwargs)

    # FIXME: the build method is only supported on AutoOperator.
    # but this doesn't distinguish if the base is an autooperator.
    # solution 1 : allow build on an Operator.
    # solution 2 : add InstanceAutoOperator.

    def build(self, **kwargs):
        # add the instance object to the kwargs
        d = {}
        d.update(kwargs)
        d[self.base.argnames[0]] = self.instance
        return self.base.build(**d)

    def __getattr__(self, attrname):
        return getattr(self.base, attrname)

def _to_ordereddict(ain):

    # convert ain to an ordered dict,
    # supported syntax:
    # [('a', '*'), ('b', '*')]
    # ['a', ('b', '*')]
    # ['a', 'b']
    # or a dictionary

    if isinstance(ain, (list, tuple)):
        ain = [
            a if isinstance(a, tuple) else (a, '*') for a in ain
        ]
    elif not isinstance(ain, (dict, OrderedDict)):
        # like a scalar, assuming it is string?
        ain = [ (ain, '*') ]

    return OrderedDict(ain)

def record_copy_all(node, **kwargs):
    """ A default rcd implementation that copies all kwargs to the tape.

        the impl is used for the vjp and vjp primitives.

        the impl is can be used for the apl primitives if no rcd is given;
        but we use 'record_copy_autodiff' to save a much smaller subset
        of variables.
    """
    return kwargs

def record_copy_autodiff(node, **kwargs):
    """ A default rcd implementation that copies `useful` kwargs to the tape
        for the autodiff.

        the impl is used for the apl primitives if no rcd is given;
    """
    jvp = node.find_primitive_type('jvp')
    vjp = node.find_primitive_type('vjp')
    record = {}
    for argname, value in kwargs.items():
        if argname in jvp.argnames or argname in vjp.argnames:
            record[argname] = value

    return record

def default_jvp(node, **kwargs):
    """ Calling apl on the jvp inputs.

        This would be appropriate for many occasions.
    """

    apl = node.operator.apl
    d = {}

    # remove "_" from input variable names before calling apl
    for argname, value in kwargs.items():
        if argname.endswith('_'):
            d[argname[:-1]] = value
        else:
            d[argname] = value

    r = apl.impl(node, **d)

    if not isinstance(r, dict):
        # scalar output, no need to patch the names.
        return r
    else:
        # add '_' to the output results before returning
        r1 = {}
        for argname, value in r.items():
            if argname in apl.aout:
                r1[argname + '_'] = value
            else:
                r1[argname] = value
        return r1

def unbound(method):
    if hasattr(method, 'im_func'):
        # python 2.7 has this unbound method thing
        return method.im_func
    # python 3, cool
    return method

def operator(kls):
    """ Decorator to declare an operator object from an operator class.

        The decorator is similar to a constructor. It produces a
        new object of type Operator, and with attributes apl, jvp and vjp as
        primitives.

        An operator class must define `ain, aout` and apl, vjp, jvp functions.

        ain : dict(name => type_pattern) describes the input arguments of
              the operator
        aout : dict(name => type_pattern) describes the output arguments of
              the operator

        Currently the type_pattern is not used; the plan is to add multi-dispatch
        if it is proven to be useful.

        apl : function(self, ...) the application of the operator;
              it shall return a dictionary
              of the evaluated values (exactly the same number of aout);
              except when there is only one output argument, then
              the result can be returned directly.

              all input arguments are resolved to python objects;
              it can have extra arguments in addition to ain.
              self is the node object that is used in the model

        rcd : function(self, ...) recording the arguments for
              invoking jvp and vjp. It shall return a dict.
              the only items included in the dict are available
              to vjp and vjp; if not defined, all arguments to apl are recorded.

        jvp : function(self, ...) the jacobian vector product. The convention
              is to use '_' + argname as the name of vectors. used for back-prop.
              self is the node object that is used in the model

        vjp : function(self, ...) the vector jacobian product. The convention
              is to use argname + '_' as the name of vectors. used for foward-prop.
              self is the node object that is used in the model

    """

    opr = Operator(kls)

    return opr

def zerobypass(impl):
    def zerobypassimpl(__node__, **kwargs):
        self = __node__
        ain = self.primitive.ain
        aout = self.primitive.aout
        if all(ZeroGradient.is_zero(kwargs[argname]) for argname in ain):
            d = {}
            for argname in aout:
                d[argname] = ZeroGradient
            return d
        return impl(self, **kwargs)
    zerobypassimpl.original = impl
    return zerobypassimpl

def _make_primitive(opr, func, impl, argnames=None, outnames=None, record_impl=record_copy_all):
    """ create primitives for the operator.

        This is used to define a primitive based on the unbound method
        defined in the operator class.

    """
    from .primitive import Primitive

    assert func in ('apl', 'vjp', 'jvp')

    aout = OrderedDict()
    ain = OrderedDict()
    if argnames is None:
        argnames = impl.__code__.co_varnames[1:impl.__code__.co_argcount]

    if func == 'apl':
        ain = opr.ain
        aout = opr.aout
    elif func == 'vjp' : # in and out are prefixed.
        for arg in opr.ain:
            aout['_' + arg] = opr.ain[arg]

        for arg in opr.aout:
            # skip unused input vector args
            if '_' + arg not in argnames: continue
            ain['_' + arg] = opr.aout[arg]
        impl = zerobypass(impl)

    elif func == 'jvp' : # in and out are prefixed.
        for arg in opr.ain:
            # skip unused input vector args
            if arg + '_' not in argnames: continue
            ain[arg + '_'] = opr.ain[arg]

        for arg in opr.aout:
            aout[arg + '_'] = opr.aout[arg]
        impl = zerobypass(impl)

    if outnames is None:
        outnames = list(aout.keys())

    primitive = Primitive(func, opr, ain, aout, argnames, outnames, impl, record_impl)

    return primitive




from __future__ import print_function

from vmad.core.model import Builder
from vmad.core.error import BadArgument
from vmad.core.autooperator import autooperator
from vmad.lib.linalg import add
import pytest

@autooperator
class example:
    ain  = {'x': '*'}
    aout = {'y': '*'}

    # must take both extra parameters and input parameters
    def main(self, x, n):
        for i in range(n):
            x = add(x1=x, x2=x)
        return dict(y=x)

@autooperator('x->y')
def example_func(x, n):
    for i in range(n):
        x = add(x1=x, x2=x)
    return dict(y=x)

def test_autooperator_annotations():
    @autooperator
    def example_func(x :'*', n) -> 'y':
        return dict(y=x)
    assert 'x' in example_func.ain
    assert 'n' not in example_func.ain
    assert 'y' in example_func.aout

    y = example_func.build(n=2).compute(vout='y', init=dict(x=1))

def test_autooperator_explicit():
    @autooperator('x', '->', 'y')
    def example_func(x, n):
        return dict(y=x)
    assert 'x' in example_func.ain
    assert 'n' not in example_func.ain
    assert 'y' in example_func.aout

    y = example_func.build(n=2).compute(vout='y', init=dict(x=1))

def test_autooperator_explicit_short():
    @autooperator('x->y')
    def example_func(x, n):
        return dict(y=x)

    assert 'x' in example_func.ain
    assert 'n' not in example_func.ain
    assert 'y' in example_func.aout

    y = example_func.build(n=2).compute(vout='y', init=dict(x=1))

def test_autooperator_explicit_short_space():
    @autooperator(' x ->y')
    def example_func(x, n):
        return dict(y=x)

    assert 'x' in example_func.ain
    assert 'n' not in example_func.ain
    assert 'y' in example_func.aout

    y = example_func.build(n=2).compute(vout='y', init=dict(x=1))

def test_autooperator_as_member():
    class MyType(object):
        def __init__(self):
            self.n = 3

        @autooperator('x->y')
        def example_func(self, x):
            return dict(y=x * self.n)

    obj = MyType()
    with Builder() as m:
        a = m.input('a')
        b = obj.example_func(a)
        m.output(b=b)

    y = m.compute(vout='b', init=dict(a=1))
    assert y == 3

    y = obj.example_func.build().compute(vout='y', init=dict(x=1.))
    assert y == 3

def test_model_nested():

    with Builder() as m:
        a = m.input('a')
        b = example(a, 2)
        c = example.forgetful(a, 2)
        m.output(b=b, c=c)

    init = dict(a = 1.0)
    (b, c), tape = m.compute(init=init, vout=['b', 'c'], monitor=print, return_tape=True)

    assert b == 4.0
    assert c == 4.0

    vjp = tape.get_vjp()
    init = dict(_b = 1.0, _c=1.0)
    _a = vjp.compute(init=init, vout='_a', monitor=print)
    assert _a == 8.0

    jvp = tape.get_jvp()
    init = dict(a_ = 1.0)
    b_, c_ = jvp.compute(init=init, vout=['b_', 'c_'], monitor=print)
    assert b_ == 4.0
    assert c_ == 4.0

def test_autooperator_additional_hyper():
    @autooperator
    def example_func(x :'*') -> 'y':
        return x

    assert 'x' in example_func.ain
    assert 'y' in example_func.aout

    @autooperator
    def main1(x : '*', n) -> 'y':
        return example_func(x, n=n)

    with pytest.raises(BadArgument):
        y = main1.build(n=2).compute(vout='y', init=dict(x=1))

def test_model_nested_build():

    m = example.build(n=2)
    init = dict(x = 1.0)
    y, tape = m.compute(init=init, vout='y', monitor=print, return_tape=True)
    assert y == 4.0

def test_autooperator_bind():

    op1 = example.bind(n=2)
    m = example.build(n=2)

    assert 'n' in op1.hyperargs
    assert 'n' not in example.hyperargs

    with Builder() as m:
        a = m.input('a')
        b = example(a, 2)
        c = op1(a)
        m.output(b=b, c=c)

    init = dict(a = 1.0)
    (b,c), tape = m.compute(init=init, vout=['b', 'c'], monitor=print, return_tape=True)

    assert b == c
    assert b == 4.0

    assert example.__bound_model__ is None
    assert op1.__bound_model__ is not None

    op1.build()

def test_autooperator_compute():

    y = example.build(n=2).compute(vout='y', init=dict(x=1))

def test_autooperator_forgetful_compute():

    y = example.forgetful.build(n=2).compute(vout='y', init=dict(x=1))

def test_autooperator_precompute():

    op1 = example.precompute(n=2, x=1)
    m = example.build(n=2)

    with Builder() as m:
        a = m.input('a')
        b = example(a, 2)
        c = op1(a, n=2)
        m.output(b=b, c=c)

    init = dict(a = 1.0)
    (b,c), tape = m.compute(init=init, vout=['b', 'c'], monitor=print, return_tape=True)

    assert b == c
    assert b == 4.0

    assert example.__bound_tape__ is None
    assert op1.__bound_tape__ is not None

    op1.build()

def test_autooperator_precompute2():

    op1 = example_func.precompute(n=2, x=1)
    m = example_func.build(n=2)

    with Builder() as m:
        a = m.input('a')
        b = example_func(a, 2)
        c = op1(a, n=2)
        m.output(b=b, c=c)

    init = dict(a = 1.0)
    (b,c), tape = m.compute(init=init, vout=['b', 'c'], monitor=print, return_tape=True)

    assert b == c
    assert b == 4.0

    assert example_func.__bound_tape__ is None
    assert op1.__bound_tape__ is not None

    op1.build()




from __future__ import print_function
from pprint import pprint
from vmad.lib.linalg import mul
from vmad.lib.linalg import add
from vmad.core.model import Builder
import pytest

def test_model_build():
    from vmad import autooperator, operator
    from vmad.core.symbol import Literal

    @autooperator('Sl->Xh')
    def upsample2(Sl, Ql):  #pm has the same resolution as Ql

        layout = add(Ql, 1)
        print(id(layout._model), len(layout._model))
        Ql1 = add(Ql, layout)
        print(id(layout._model), id(Ql1._model), len(layout._model))
        Sl1 = add(Sl, layout)
        print(id(layout._model), id(Ql1._model), len(layout._model))

        dis_d = add(Ql1, Sl1)
        print(id(layout._model), id(Ql1._model), len(layout._model))

        return dis_d

    Sl = 1
    model2 = upsample2.build(Ql=Sl)

    Xh = model2.compute('Xh', init=dict(Sl=Sl))
    (y, ), [vjp] = model2.compute_with_vjp(init=dict(Sl=Sl), v=dict(_Xh=1))

def test_model_partial():
    with Builder() as m:
        a = m.input('a')
        t1 = add(x1=a, x2=a)
        m.output(c=t1)

    init = dict(a=3)

    c, tape = m.compute(init=init, vout='c', return_tape=True)
    assert c == 6

    vjp = tape.get_vjp()
    init = dict(_c=1.0)
    _a = vjp.compute(init=init, vout='_a', monitor=print)
    assert _a == 2.0

    jvp = tape.get_jvp()
    init = dict(a_=1.0)
    c_ = jvp.compute(init=init, vout='c_', monitor=print)
    assert c_ == 2.0

def test_model_unused():
    with Builder() as m:
        a, b = m.input('a', 'b')
        # use a twice with a dependency
        # triggers problem with last_ref in autodiff;
        # because this line is not executed by the context;
        # last_ref is not True for the last ref on the tape.
        d = (a + b) + a
        m.output(c=1.0)

    init = dict(a=3, b=4)
    c, tape = m.compute(init=init, vout='c', return_tape=True)
    assert c == 1.0

    vjp = tape.get_vjp()
    init = dict(_c=1.0)
    _a, _b = vjp.compute(init=init, vout=['_a', '_b'], monitor=print)
    assert _a == 0
    assert _b == 0

    jvp = tape.get_jvp()
    init = dict(a_=1.0, b_=1.0)
    c_ = jvp.compute(init=init, vout='c_', monitor=print)
    assert c_ == 0

def test_model_partial_out():
    with Builder() as m:
        a = m.input('a')
        t1 = add(x1=a, x2=a)
        m.output(c=t1)
        m.output(a=a)

    init = dict(a=3)

    (a, c), tape = m.compute(init=init, vout=['a', 'c'], return_tape=True)
    assert c == 6
    assert a == 3

    vjp = tape.get_vjp()

    # test two outputs individually
    init = dict(_c=1.0, _a=0.0)
    _a = vjp.compute(init=init, vout='_a', monitor=print)
    assert _a == 2.0

    init = dict(_c=0.0, _a=1.0)
    _a = vjp.compute(init=init, vout='_a', monitor=print)
    assert _a == 1.0

    jvp = tape.get_jvp()
    init = dict(a_=1.0)
    a_, c_ = jvp.compute(init=init, vout=['a_', 'c_'], monitor=print)
    assert c_ == 2.0
    assert a_ == 1.0

def test_model_many_rewrites():
    # this is a nasty model with many variable rewrites.
    n = 2
    with Builder() as m:
        x = m.input('x')
        for i in range(2):
            x = add(x1=x, x2=x)

        m.output(y=x)

    init = dict(x=1.0)
    y, tape = m.compute(init=init, vout='y', return_tape=True)
    assert y == 4.0

    vjp = tape.get_vjp()
    init = dict(_y = 1.0)
    _x = vjp.compute(init=init, vout='_x', monitor=print)
    assert _x == 4.0

def test_model_compute_with_vjp():
    with Builder() as m:
        a = m.input('a')
        t1 = add(x1=a, x2=a)
        m.output(b=t1)

    init = [('a', 1)]
    [b], [_a] = m.compute_with_vjp(init=init, v=[('_b', 1.0)])
    assert b == 2.0
    assert _a == 2.0


def test_model_compute_with_jvp():
    with Builder() as m:
        a = m.input('a')
        t1 = add(x1=a, x2=a)
        m.output(b=t1)

    init = [('a', 1)]
    b, b_ = m.compute_with_jvp(vout='b', init=init, v=[('a_', 1.0)])
    assert b == 2.0
    assert b_ == 2.0

def test_model_compute_with_gnDp():
    with Builder() as m:
        a = m.input('a')
        t1 = mul(x1=a, x2=a)
        m.output(b=t1)

    init = [('a', 1)]
    b, [_a_] = m.compute_with_gnDp(
                            vout='b',
                            init=init,
                            v=[('a_', 1.0)],
                            )
    assert b == 1.0
    assert _a_ == 4.0

def test_model_attr():
    import numpy
    with Builder() as m:
        a, b = m.input('a', 'b')
        d = add(x1=b, x2=1)
        t1 = add(x1=a, x2=b.eval(lambda b: b.size))
        m.output(c=t1)

    init = dict(a=2, b=numpy.array([2,]))

    c, tape = m.compute(init=init, vout='c', return_tape=True)
    assert c == 3

    vjp = tape.get_vjp()
    init = dict(_c=1.0)
    _a = vjp.compute(init=init, vout='_a', monitor=print)
    assert _a == 1.0

    jvp = tape.get_jvp()
    init = dict(a_=1.0)
    c_ = jvp.compute(init=init, vout='c_', monitor=print)
    assert c_ == 1.0




from vmad.core.operator import operator

class unary:
    ain = 'x'
    aout = 'y'

class binary:
    ain = 'x1', 'x2'
    aout = 'y'


@operator
class neg(unary):
    def apl(node, x):
        return dict(y = -x)

    def vjp(node, _y):
        return dict(_x = -_y)

    def jvp(node, x_):
        return dict(y_ = -x_)

@operator
class pos(unary):
    def apl(node, x):
        return dict(y = +x)

    def vjp(node, _y):
        return dict(_x = +_y)

    def jvp(node, x_):
        return dict(y_ = +x_)

@operator
class abs(unary):
    def apl(node, x):
        # function name was replaced
        from builtins import abs
        return dict(y = abs(x))

    def vjp(node, _y, x):
        return dict(_x = _y * (x > 0) + -_y * (x < 0))

    def jvp(node, x_, x):
        return dict(y_ = x_ * (x > 0) + -x_ * (x < 0))

@operator
class add(binary):
    def apl(node, x1, x2):
        return dict(y = x1 + x2)

    def vjp(node, _y):
        return dict(_x1 = _y, _x2 = _y)

    def jvp(node, x1_, x2_):
        return dict(y_ = x1_ + x2_)

@operator
class sub(binary):
    def apl(node, x1, x2):
        return dict(y = x1 - x2)

    def vjp(node, _y):
        return dict(_x1 = _y, _x2 = -_y)

    def jvp(node, x1_, x2_):
        return dict(y_ = x1_ + x2_)

@operator
class mul(binary):
    def apl(node, x1, x2):
        return dict(y = x1 * x2)

    def rcd(node, x1, x2, y):
        # the other value is not needed, 0 should work.
        if node.is_literal('x1'):
            x2 = 0
        if node.is_literal('x2'):
            x1 = 0
        return dict(x1=x1, x2=x2)

    def vjp(node, _y, x1, x2):
        return dict(_x1 = _y * x2,
                    _x2 = _y * x1)

    def jvp(node, x1_, x2_, x1, x2):
        return dict(y_ = x1_* x2 + x1 * x2_)

@operator
class div(binary):
    def apl(node, x1, x2):
        x2inv = 1 / x2
        return dict(y = x1 * x2inv, x2inv=x2inv)

    def rcd(node, x1, x2inv, y, x2):
        from vmad.core.symbol import Literal
        # the other value is not needed, 0 should work.
        if node.is_literal('x1'):
            x1fac = 0
        else:
            x1fac = 1 / x2
        if node.is_literal('x2'):
            x2fac = 0
        else:
            x2fac = x1 * (1 / x2) * (1 / x2)
        return dict(x1fac=x1fac, x2fac=x2fac)

    def vjp(node, _y, x1fac, x2fac):
        return dict(_x1 = _y * x1fac,
                    _x2 = -_y * x2fac)

    def jvp(node, x1_, x2_, x1fac, x2fac):
        return dict(y_ = x1_ * x1fac - x2fac * x2_)

@operator
class mod(binary):
    def apl(node, x1, x2):
        return dict(y = x1 % x2, n = x1 // x2)

    def vjp(node, _y, x1, x2, n):
        return dict(_x1 = _y,
                    _x2 = (- n) * _y)

    def jvp(node, x1_, x2_, x1, x2, n):
        return dict(y_ = x1_ - n * x2_)

@operator
class pow(binary):
    def apl(node, x1, x2):
        if not node.is_literal('x2'):
            from numpy import log
            logx1 = log(x1)
        else:
            # no need for logx2, as _x2 will be zeros.
            logx1 = 0

        return dict(y=x1 ** x2, logx1=logx1)

    def vjp(node, _y, x1, x2, logx1):
        fac = x1 ** (x2 - 1) if x2 != 1 else 1
        return dict(_x1 = x2 * _y * fac,
                    _x2 = _y * x1**x2 * logx1)

    def jvp(node, x1_, x2_, x1, x2, logx1):
        fac = x1 ** (x2 - 1) if x2 != 1 else 1
        return dict(y_ = x2 * x1_ * fac + x2_ * x1**x2 * logx1)





"""
    This module provides utils for debugging a model.

    These functions are triggered during model execution.
"""

from vmad.core.operator import operator, ZeroGradient

@operator
class eval:
    ain  = 'x'
    aout = 'y'

    def apl(node, x, function):
        """ Evaluates a function on the symbol x at run time.

            Parameters
            ----------
            function : callable or string.
                if a string is provided for the evaluation, x is the variable.
                if a function is provided, it shall be function(x)

        """
        if hasattr(function, '__call__'):
            return dict(y=function(x))
        else:
            from builtins import eval
            return dict(y=eval(function, dict(x=x)))

    def vjp(node): return ZeroGradient
    def jvp(node): return ZeroGradient

@operator
class watchpoint:
    """ printing a variable or a list of variables.

        >>> with Builder() as m:
        >>>    x = m.input('x')
        >>>    watchpoint(x, lambda x: print(x.shape))
        >>>    watchpoint(x, lambda x: print(x.size))
        >>>    watchpoint((x, x), lambda (x1, x2): print(x1.size))
        >>>    m.output(x)
    """
    ain  = {'x': '*'}
    aout = {}

    def apl(node, x, monitor=print):
        monitor(x)
        return dict(y = x)

    def vjp(node): return ZeroGradient
    def jvp(node): return

@operator
class assert_isinstance:
    """ assert a variable is of a certain type """
    ain = 'obj'
    aout = []
    def apl(node, obj, class_or_tuple):
        if not isinstance(obj, class_or_tuple):
            raise TypeError('Expecting an instance of %s, got %s', repr(class_or_tuple), repr(type(obj)))

    def vjp(node): return ZeroGradient
    def jvp(node): return ZeroGradient

@operator
class assert_true:
    """ assert if a function applied to a variable is true.

        Examples:

        >>> with Builder() as m:
        >>>    x = m.input('x')
        >>>    assert_true(x, lambda x: isinstance(x, int)))
        >>>    m.output(x)
    """

    ain = 'x'
    aout = []
    def apl(node, x, func):
        if not func(x):
            raise AssertionError('Assertion failed on %s(%s).' % (func, x))

    def vjp(node): return ZeroGradient
    def jvp(node): return ZeroGradient



from __future__ import print_function

from pprint import pprint
from vmad.core.model import Builder
import pytest
from vmad.lib import linalg

def test_operator_watchpoint():
    from vmad.core.stdlib import watchpoint
    foo = [0]
    def monitor(x):
        foo[0] = 1

    with Builder() as m:
        a = m.input('a')
        b = linalg.add(a, a)
        watchpoint(a, monitor=monitor)
        m.output(c=b)

    init = [('a', 1)]
#    for node in m:
#        print('m', node)
    c, tape = m.compute(init=init, vout='c', return_tape=True)
#    vjp = tape.get_vjp()
#    for node in vjp:
#        print('vjp', node)

    [c], [_a] = m.compute_with_vjp(init=init, v=[('_c', 1.0)])
    assert foo[0] == 1
    assert c == 2
    assert _a == 2.0

    foo[0] = 0
    c, c_ = m.compute_with_jvp(vout='c', init=init, v=[('a_', 1.0)])
    assert foo[0] == 1
    assert c == 2
    assert c_ == 2.0


def test_operator_assert_isinstance():
    from vmad.core.stdlib import assert_isinstance

    with Builder() as m:
        a = m.input('a')
        assert_isinstance(a, int)
        m.output(c=a)

    init = [('a', 1)]
    c, tape = m.compute(init=init, return_tape=True, vout='c')

    [c], [_a] = m.compute_with_vjp(init=init, v=[('_c', 1.0)])
    assert c == 1
    assert _a == 1.0

    c, c_ = m.compute_with_jvp(vout='c', init=init, v=[('a_', 1.0)])
    assert c == 1
    assert c_ == 1.0

    with pytest.raises(TypeError):
        c = m.compute(vout='c', init=dict(a=1.09))

def test_operator_assert_true():
    from vmad.core.stdlib import assert_true

    with Builder() as m:
        a = m.input('a')
        assert_true(a, lambda x: isinstance(x, int))
        m.output(c=a)

    with pytest.raises(AssertionError):
        c = m.compute(vout='c', init=dict(a=1.09))

    c = m.compute(vout='c', init=dict(a=1))

def test_div_error():
    from vmad import autooperator
    @autooperator('a->c')
    def test(a, b):
        c= b/a
        return c
    model = test.build(b=3)
    (y, ), (vjp, ) = model.compute_with_vjp(init=dict(a=2), v=dict(_c=1.0))
    assert y == 1.5
    assert vjp == -0.75



from __future__ import print_function
from pprint import pprint
from vmad.lib import linalg
import numpy

from vmad.lib import binary

from vmad import Builder
from vmad.testing import BaseVectorTest

class BinaryUfuncVectorTest(BaseVectorTest):
    ufunc = None

    def y(self, x):
        return type(self).ufunc.prototype.f(x[0], x[1])

    def model(self, x):
        x1 = linalg.take(x, 0, axis=0)
        x2 = linalg.take(x, 1, axis=0)
        return type(self).ufunc(x1, x2)


class Test_fmax(BinaryUfuncVectorTest):
    ufunc = binary.fmax
    x = numpy.arange(10).reshape(2, 5)

class Test_fmin(BinaryUfuncVectorTest):
    ufunc = binary.fmin
    x = numpy.arange(10).reshape(2, 5)

class Test_add(BinaryUfuncVectorTest):
    ufunc = binary.add
    x = numpy.arange(10).reshape(2, 5)




from .error import DuplicatedOutput, InferError
from .context import Context
from .tape import Tape
from .symbol import Symbol
from .stdlib import terminal

from collections import OrderedDict
import uuid
import weakref

class Model(list):
    def __init__(self):
        self._vin = []
        self._vout = []

        self._name = uuid.uuid4().hex
        self._symbols = set()

    def __hash__(self):
        return hash(self._name)

    def __eq__(self, other):
        # needed in making sure model discovery is unique.
        return self._name == other._name

    def extend(self, other):
        """ concatenate the model with another model.
        """

        # _syms are merged as we reanchor the _model of
        # symbols.

        self._vin.extend(other._vin)
        self._vout.extend(other._vout)

        for symbol in other._symbols:
            self.anchor(symbol)

        list.extend(self, other)

    def input(self, *args):
        """ Declare multiple input variables.

            Parameters
            ----------
            args : list of input variables.
                    can be either instances of Symbol or
                    a string, in which case a Symbol is
                    created.

            Returns
            -------
            variables : list or a single object if len(args) == 1
                Symbols, marked as inputs and anchored to the model.

        """
        # FIXME: assert the Symbol hasn't been used
        v = []
        for name in args:
            if isinstance(name, Symbol):
                r = name
            else:
                r = Symbol(name)

            # anchor it to self
            self.anchor(r)
            v.append(r)

        self._vin.extend(v)

        if len(args) == 1:
            return v[0]

        return v

    def anchor(self, symbol):
        if symbol._anchored:
            if not symbol._transient:
                if symbol._model is not self:
                    raise ModelError("cannot change the model after the symbol is no longer in transient.")

        if isinstance(self, ModelInTransient):
            # use strong reference to keep these Transient models alive.
            # they will be replaced by the Model in the end as we put symbols
            # together via primitives.
            symbol._model_ref = self
        else:
            symbol._model_ref = weakref.ref(self)

        self._symbols.add(symbol)

    def output(self, **kwargs):
        for varname, oldvar in kwargs.items():
            for var in self._vout:
                if var._name == varname:
                    raise DuplicatedOutput("Variable %s is already marked as an output" % varname)

            var = Symbol(varname, model=self)
            terminal.apl.create_node(dict(x=oldvar), dict(y=var))
            self._vout.append(var)

    def compile(self): pass

    def unique_name(self, str):
        return '%s-%s' % (str, uuid.uuid4().hex)

    def __repr__(self):
        return "Model: %s => %s" % (self._vin, self._vout)

    def compute(self, vout, init, return_tape=False, return_dict=False, monitor=None):
        """
            compute a model with the initial values

            init : dictionary
        """
        if isinstance(vout, str):
            single_return = True
            vout = [vout]
        else:
            single_return = False

        init = OrderedDict(init)

        if return_tape:
            tape = Tape(self, init)
        else:
            tape = None

        ctx = Context(**init)

        out = ctx.compute(self, vout=vout,
                        tape=tape,
                        monitor=monitor)

        if return_tape:
            tape.finalize(out)

        if not return_dict:
            out = list(out.values())
            if single_return:
                out = out[0]

        if return_tape:
            out = out, tape

        return out

    def compute_with_vjp(self, init, v, return_dict=False, monitor=None):
        """
            compute a model with the vjp

            Parameters
            ----------
            init : dict, or list of tuples
                initial values of the module. List will
                be convert to an ordereddict
            v    : dict, or list of tuples
                initial vectors dotted to the jacobian. List will
                be convert to an ordereddict

            The vout of the model computation is inferred
            from variable names of v.

            The vout of the vjp computation is inferred from
            variable names of init.

            Returns
            -------
            out : a list of computed values of the model;
                ordered by the keys of the dictionary.

            vjpout : a list of computed values of the vjp
                ordered by the keys of the dictionary.

        """

        init = OrderedDict(init)
        v = OrderedDict(v)

        # assert all of the vectors start with '_', such that
        # we can properly infer the vout of the model
        assert all(varname.startswith('_') for varname in v.keys())

        vout = [varname[1:] for varname in v.keys()]
        out, tape = self.compute(vout, init=init, return_tape=True)
        vjpvout = tape.get_vjp_vout()

        vjp = tape.get_vjp()
        vjpout = vjp.compute(vjpvout, init=v)

        if return_dict:
            out = OrderedDict(zip(vout, out))
            vjpout = OrderedDict(zip(vjpvout, vjpout))

        return out, vjpout

    def compute_with_jvp(self, vout, init, v, return_dict=False, monitor=None):
        """
            compute a model with the jvp

            Parameters
            ----------
            vout : list or string
                output variable to compute.
            init : dict, or list of tuples
                initial values of the module.
            v    : dict, or list of tuples
                initial vectors dotted by the jacobian.

            The vout of the vjp computation is inferred from
            variable names of vout.

            Returns
            -------
            out : a list of computed values of the model.
                if vout is a string, out is delisted (a single python object)

            jvpout : a list of computed values of the jvp
                ordered by the keys of the dictionary.
                if vout is a string, out is delisted (a single python object)

        """
        init = OrderedDict(init)
        v = OrderedDict(v)

        # assert all of the vectors ends with '_'.
        assert all(varname.endswith('_') for varname in v.keys())

        out, tape = self.compute(vout, init=init, return_tape=True)

        jvpvout = tape.get_jvp_vout()

        jvp = tape.get_jvp()
        jvpout = jvp.compute(jvpvout, init=v)

        if return_dict:
            out = OrderedDict(zip(vout, out))
            jvpout = OrderedDict(zip(jvpvout, jvpout))
        else:
            if not isinstance(vout, (tuple, list)):
                jvpout = jvpout[0]

        return out, jvpout

    def compute_with_gnDp(self, vout, init, v, return_dict=False, monitor=None):
        """
            compute a model with the gauss-newton approximated
            hessian vector product (D dot v = JTJ dot v)

            Parameters
            ----------
            vout : list or string
                output variable to compute.
            init : dict, or list of tuples
                initial values of the model.
            v    : dict, or list of tuples
                initial vectors dotted by the jacobian; shall
                match the variables in init, but with '_' appended.

            The vout of the vjp computation is inferred from
            variable names of vout.

            Returns
            -------
            out : a list of computed values of the model.
                if vout is a string, out is delisted (a single python object)

            gnhpout: a list of computed values of the gnhp
                ordered by the keys of the dictionary init.

        """
        init = OrderedDict(init)
        v = OrderedDict(v)

        # assert all of the vectors ends with '_'.
        assert all(varname.endswith('_') for varname in v.keys())

        out, tape = self.compute(vout, init=init, return_tape=True)

        jvp = tape.get_jvp()
        jvpvout = tape.get_jvp_vout()
        jvpout = jvp.compute(jvpvout, init=v)

        vjp = tape.get_vjp()
        vjpvout = tape.get_vjp_vout()
        # connecting the output of jvp as input of vjp
        vjpvin = ['_' + i[:-1] for i in jvpvout]
        gnhpout = vjp.compute(vjpvout, init=zip(vjpvin, jvpout))

        if return_dict:
            out = OrderedDict(zip(vout, out))
            gnhpout = OrderedDict(zip(vjpvout, gnhpout))

        return out, gnhpout

    @staticmethod
    def consolidate(models):
        """ Consolidate several models into a single one.

            The non-transient model is preferred.

            After this call all but the returned model
            shall be considered invalid.
        """
        model = None
        # find the non-transient model
        for i in models:
            if isinstance(i, ModelInTransient): continue
            if i is None: continue
            if model is None: # first time seeing a non transient
                model = i
            else:
                raise InferError("Multiple non transient models are found")

        if model is None:
            if len(models) == 0:
                model = ModelInTransient()
            else:
                # get the first model
                model = next(iter(models))

        for i in models:
            if i is model: continue # skip 'the' model
            model.extend(i)

        return model

class ModelInTransient(Model):
    """ A model that is in transient, eventually will be added to a 'real' model.

        This is necessary as we start building a model with symbols that haven't been
        associated to the desired model.

        They will be first added to a transient model,
        then when the true model is seen they will be added to the model.
    """
    pass

class Builder(Model):
    """ A context manager to signify the process of buildig a model.

    """
    def __enter__(self): return self
    def __exit__(self, *args): self.compile()



from __future__ import print_function
from pprint import pprint
from vmad.lib import linalg
import numpy

from vmad.lib import unary

from vmad import Builder
from vmad.testing import BaseVectorTest

class UnaryUfuncVectorTest(BaseVectorTest):
    ufunc = None

    def y(self, x):
        return type(self).ufunc.prototype.f(x)

    def model(self, x):
        return type(self).ufunc(x)

class Test_sinc(UnaryUfuncVectorTest):
    ufunc = unary.sinc
    x = numpy.arange(-10, 10)


class Test_sin(UnaryUfuncVectorTest):
    ufunc = unary.sin
    x = numpy.arange(-10, 10)

class Test_cos(UnaryUfuncVectorTest):
    ufunc = unary.cos
    x = numpy.arange(-10, 10)

class Test_tan(UnaryUfuncVectorTest):
    ufunc = unary.tan
    x = numpy.linspace(-1, 1, 10)

class Test_arcsin(UnaryUfuncVectorTest):
    ufunc = unary.arcsin
    x = numpy.linspace(-0.99, 0.99, 8)

class Test_arccos(UnaryUfuncVectorTest):
    ufunc = unary.arccos
    x = numpy.linspace(-0.99, 0.99, 8)

class Test_arctan(UnaryUfuncVectorTest):
    ufunc = unary.arctan
    x = numpy.arange(-10, 10)

class Test_log(UnaryUfuncVectorTest):
    ufunc = unary.log
    x = numpy.arange(10) + 1

class Test_log10(UnaryUfuncVectorTest):
    ufunc = unary.log10
    x = numpy.arange(10) + 1

class Test_exp(UnaryUfuncVectorTest):
    ufunc = unary.exp
    x = numpy.linspace(-0.99, 0.99, 8)

class Test_absolute(UnaryUfuncVectorTest):
    ufunc = unary.absolute
    x = numpy.linspace(-0.99, 0.99, 8)

class Test_sinh(UnaryUfuncVectorTest):
    ufunc = unary.sinh
    x = numpy.linspace(-0.99, 0.99, 8)

class Test_cosh(UnaryUfuncVectorTest):
    ufunc = unary.cosh
    x = numpy.linspace(-0.99, 0.99, 8)



from vmad import operator
from vmad.core.symbol import Literal, ZeroLiteral
from vmad.core.symbol import Symbol
import numpy

class binary_ufunc:
    ain = 'x1', 'x2'
    aout = 'y'

    @staticmethod
    def f(x1, x2): return x1

    @staticmethod
    def fprime1(x1, x2): return 1

    @staticmethod
    def fprime2(x1, x2): return 0

    def apl(node, x1, x2):
        return node.operator.prototype.f(x1, x2)

    def vjp(node, x1, x2, _y):
        return dict(
                _x1 = node.operator.prototype.fprime1(x1, x2) * _y,
                _x2 = node.operator.prototype.fprime2(x1, x2) * _y
                )

    def jvp(node, x1, x2, x1_, x2_):
        return node.operator.prototype.fprime1(x1, x2) * x1_ \
            +  node.operator.prototype.fprime2(x1, x2) * x2_

@operator
class fmax(binary_ufunc):
    @staticmethod
    def f(x1, x2):
        return numpy.fmax(x1, x2)

    @staticmethod
    def fprime1(x1, x2):
        return x1 >= x2

    @staticmethod
    def fprime2(x1, x2):
        return ~(x1 >= x2)

maximum = fmax

@operator
class fmin(binary_ufunc):
    @staticmethod
    def f(x1, x2):
        return numpy.fmin(x1, x2)

    @staticmethod
    def fprime1(x1, x2):
        return x1 < x2

    @staticmethod
    def fprime2(x1, x2):
        return ~(x1 <= x2)

minimum = fmax

@operator
class add(binary_ufunc):
    @staticmethod
    def f(x1, x2):
        return numpy.add(x1, x2)

    @staticmethod
    def fprime1(x1, x2):
        return 1

    @staticmethod
    def fprime2(x1, x2):
        return 1

